# Copyright 2017 VMware, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import netaddr

from neutron_dynamic_routing.extensions import bgp as bgp_ext
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from neutron_lib.api.definitions import address_scope
from neutron_lib.api.definitions import external_net as extnet_apidef
from neutron_lib import constants as n_const
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory
from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.common import locking
from vmware_nsx.common import nsxv_constants
from vmware_nsx.db import nsxv_db
from vmware_nsx.extensions import edge_service_gateway_bgp_peer as ext_esg_peer
from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.plugins.nsx_v.vshield.common import exceptions as vcns_exc

LOG = logging.getLogger(__name__)


def ip_prefix(name, ip_address):
    return {'ipPrefix': {'name': name, 'ipAddress': ip_address}}


def redistribution_rule(advertise_static_routes, prefix_name, action='permit'):
    rule = {
        'prefixName': prefix_name,
        'action': action,
        'from': {
            'ospf': False,
            'bgp': False,
            'connected': not advertise_static_routes,
            'static': advertise_static_routes
        }
    }
    return {'rule': rule}


def _get_bgp_neighbour(ip_address, remote_as, password, direction):
    bgp_filter = {'bgpFilter': [{'direction': direction, 'action': 'permit'}]}
    nbr = {
        'ipAddress': ip_address,
        'remoteAS': remote_as,
        'bgpFilters': bgp_filter,
        'holdDownTimer': cfg.CONF.nsxv.bgp_neighbour_hold_down_timer,
        'keepAliveTimer': cfg.CONF.nsxv.bgp_neighbour_keep_alive_timer
    }
    if password:
        nbr['password'] = password
    return {'bgpNeighbour': nbr}


def bgp_neighbour_from_peer(bgp_peer):
    return _get_bgp_neighbour(bgp_peer['peer_ip'],
                              bgp_peer['remote_as'],
                              bgp_peer['password'],
                              direction='out')


def gw_bgp_neighbour(ip_address, remote_as, password):
    return _get_bgp_neighbour(ip_address, remote_as, password,
                              direction='in')


class NSXvBgpDriver(object):
    """Class driver to address the neutron_dynamic_routing API"""

    def __init__(self, plugin):
        super(NSXvBgpDriver, self).__init__()
        self._plugin = plugin
        self._core_plugin = directory.get_plugin()
        if self._core_plugin.is_tvd_plugin():
            self._core_plugin = self._core_plugin.get_plugin_by_type(
                projectpluginmap.NsxPlugins.NSX_V)
        if not self._core_plugin:
            err_msg = _("NSXv BGP cannot work without the NSX-V core plugin")
            raise n_exc.InvalidInput(error_message=err_msg)
        self._nsxv = self._core_plugin.nsx_v
        self._edge_manager = self._core_plugin.edge_manager

    def prefix_name(self, subnet_id):
        return 'subnet-%s' % subnet_id

    def _get_router_edge_info(self, context, router_id):
        edge_binding = nsxv_db.get_nsxv_router_binding(context.session,
                                                       router_id)
        if not edge_binding:
            return None, None

        # Indicates which routes should be advertised - connected or static.
        advertise_static_routes = False
        if edge_binding['edge_type'] != nsxv_constants.SERVICE_EDGE:
            # Distributed router
            plr_id = self._edge_manager.get_plr_by_tlr_id(context, router_id)
            edge_binding = nsxv_db.get_nsxv_router_binding(context.session,
                                                           plr_id)
            if not edge_binding:
                # Distributed router isn't bound to plr
                return None, None
            # PLR for distributed router, advertise static routes.
            advertise_static_routes = True
        return edge_binding['edge_id'], advertise_static_routes

    def get_advertised_routes(self, context, bgp_speaker_id):
        routes = []
        bgp_speaker = self._plugin.get_bgp_speaker(context, bgp_speaker_id)
        edge_router_dict = (
            self._get_dynamic_routing_edge_list(context,
                                                bgp_speaker['networks'][0],
                                                bgp_speaker_id))
        for edge_id, edge_router_config in edge_router_dict.items():
            bgp_identifier = edge_router_config['bgp_identifier']
            subnets = self._query_tenant_subnets(
                context, edge_router_config['no_snat_routers'])
            routes.extend([(subnet['cidr'], bgp_identifier)
                           for subnet in subnets])
        routes = self._plugin._make_advertised_routes_list(routes)
        return self._plugin._make_advertised_routes_dict(routes)

    def _get_dynamic_routing_edge_list(self, context,
                                       gateway_network_id, bgp_speaker_id):
        # Filter the routers attached this network as gateway interface
        filters = {'network_id': [gateway_network_id],
                   'device_owner': [n_const.DEVICE_OWNER_ROUTER_GW]}
        fields = ['device_id', 'fixed_ips']
        gateway_ports = self._core_plugin.get_ports(context, filters=filters,
                                                    fields=fields)

        bgp_bindings = nsxv_db.get_nsxv_bgp_speaker_bindings(
            context.session, bgp_speaker_id)
        binding_info = {bgp_binding['edge_id']: bgp_binding['bgp_identifier']
                        for bgp_binding in bgp_bindings}

        edge_router_dict = {}
        for port in gateway_ports:
            router_id = port['device_id']
            router = self._core_plugin._get_router(context, router_id)
            edge_id, advertise_static_routes = (
                self._get_router_edge_info(context, router_id))
            if not edge_id:
                # Shared router is not attached on any edge
                continue

            if edge_id not in edge_router_dict:
                bgp_identifier = binding_info.get(
                    edge_id, port['fixed_ips'][0]['ip_address'])
                edge_router_dict[edge_id] = {'no_snat_routers': [],
                                             'bgp_identifier':
                                             bgp_identifier,
                                             'advertise_static_routes':
                                             advertise_static_routes}
            if not router.enable_snat:
                edge_router_dict[edge_id]['no_snat_routers'].append(router_id)
        return edge_router_dict

    def _get_md_proxy_for_router(self, context, router_id):
        binding = nsxv_db.get_nsxv_router_binding(context.session,
                                                  router_id)
        md_proxy = None
        if binding:
            az_name = binding['availability_zone']
            md_proxy = self._core_plugin.get_metadata_proxy_handler(
                az_name)
        return md_proxy

    def _query_tenant_subnets(self, context, router_ids):
        # Query subnets attached to all of routers attached to same edge
        subnets = []
        for router_id in router_ids:
            filters = {'device_id': [router_id],
                       'device_owner': [n_const.DEVICE_OWNER_ROUTER_INTF]}
            int_ports = self._core_plugin.get_ports(context,
                                                    filters=filters,
                                                    fields=['fixed_ips'])
            # We need to skip metadata subnets
            md_proxy = self._get_md_proxy_for_router(context, router_id)
            for p in int_ports:
                subnet_id = p['fixed_ips'][0]['subnet_id']
                if md_proxy and md_proxy.is_md_subnet(subnet_id):
                    continue
                subnet = self._core_plugin.get_subnet(context, subnet_id)
                subnets.append({'id': subnet_id,
                                'cidr': subnet['cidr']})
        LOG.debug("Got related subnets %s", subnets)
        return subnets

    def _get_bgp_speakers_by_bgp_peer(self, context, bgp_peer_id):
        fields = ['id', 'peers']
        bgp_speakers = self._plugin.get_bgp_speakers(context, fields=fields)
        bgp_speaker_ids = [bgp_speaker['id'] for bgp_speaker in bgp_speakers
                           if bgp_peer_id in bgp_speaker['peers']]
        return bgp_speaker_ids

    def _get_prefixes_and_redistribution_rules(self, subnets,
                                               advertise_static_routes):
        prefixes = []
        redis_rules = []
        for subnet in subnets:
            prefix_name = self.prefix_name(subnet['id'])
            prefix = ip_prefix(prefix_name, subnet['cidr'])
            prefixes.append(prefix)
            rule = redistribution_rule(advertise_static_routes, prefix_name)
            redis_rules.append(rule)
        return prefixes, redis_rules

    def create_bgp_speaker(self, context, bgp_speaker):
        bgp_speaker_data = bgp_speaker['bgp_speaker']
        ip_version = bgp_speaker_data.get('ip_version')
        if ip_version and ip_version == 6:
            err_msg = _("NSXv BGP does not support for IPv6")
            raise n_exc.InvalidInput(error_message=err_msg)

    def update_bgp_speaker(self, context, bgp_speaker_id, bgp_speaker):
        bgp_obj = bgp_speaker['bgp_speaker']
        old_speaker_info = self._plugin.get_bgp_speaker(context,
                                                        bgp_speaker_id)
        enabled_state = old_speaker_info['advertise_tenant_networks']
        new_enabled_state = bgp_obj.get('advertise_tenant_networks',
                                        enabled_state)
        if new_enabled_state == enabled_state:
            return

        bgp_bindings = nsxv_db.get_nsxv_bgp_speaker_bindings(
            context.session, bgp_speaker_id)
        edge_ids = [bgp_binding['edge_id'] for bgp_binding in bgp_bindings]
        action = 'Enabling' if new_enabled_state else 'Disabling'
        LOG.info("%s BGP route redistribution on edges: %s.", action, edge_ids)
        for edge_id in edge_ids:
            try:
                self._nsxv.update_routing_redistribution(edge_id,
                                                         new_enabled_state)
            except vcns_exc.VcnsApiException:
                LOG.warning("Failed to update BGP on edge '%s'.", edge_id)

    def delete_bgp_speaker(self, context, bgp_speaker_id):
        bgp_bindings = nsxv_db.get_nsxv_bgp_speaker_bindings(
            context.session, bgp_speaker_id)
        self._stop_bgp_on_edges(context, bgp_bindings, bgp_speaker_id)

    def _validate_bgp_configuration_on_peer_esg(self, bgp_peer):
        if not bgp_peer.get('esg_id'):
            return
        # TBD(roeyc): Validate peer_ip is on subnet

        bgp_config = self._nsxv.get_routing_bgp_config(bgp_peer['esg_id'])
        remote_as = bgp_peer['remote_as']
        esg_id = bgp_peer['esg_id']
        esg_as = bgp_config['bgp'].get('localAS')
        if not bgp_config['bgp']['enabled']:
            raise ext_esg_peer.BgpDisabledOnEsgPeer(esg_id=esg_id)
        if esg_as != int(remote_as):
            raise ext_esg_peer.EsgRemoteASDoNotMatch(remote_as=remote_as,
                                                     esg_id=esg_id,
                                                     esg_as=esg_as)
        h, resp = self._nsxv.vcns.get_interfaces(esg_id)
        for iface in resp['vnics']:
            address_groups = iface['addressGroups']['addressGroups']
            matching_iface = [ag for ag in address_groups
                              if ag['primaryAddress'] == bgp_peer['peer_ip']]
            if matching_iface:
                break
        else:
            raise ext_esg_peer.EsgInternalIfaceDoesNotMatch(esg_id=esg_id)

    def create_bgp_peer(self, context, bgp_peer):
        bgp_peer = bgp_peer['bgp_peer']
        remote_ip = bgp_peer['peer_ip']
        if not netaddr.valid_ipv4(remote_ip):
            err_msg = _("NSXv BGP does not support for IPv6")
            raise n_exc.InvalidInput(error_message=err_msg)
        self._validate_bgp_configuration_on_peer_esg(bgp_peer)

    def update_bgp_peer(self, context, bgp_peer_id, bgp_peer):
        password = bgp_peer['bgp_peer'].get('password')
        old_bgp_peer = self._plugin.get_bgp_peer(context, bgp_peer_id)

        # Only password update is relevant for backend.
        if old_bgp_peer['password'] == password:
            return

        bgp_speaker_ids = self._get_bgp_speakers_by_bgp_peer(context,
                                                             bgp_peer_id)
        # Update the password for the old bgp peer and update NSX
        old_bgp_peer['password'] = password
        neighbour = bgp_neighbour_from_peer(old_bgp_peer)
        for bgp_speaker_id in bgp_speaker_ids:
            with locking.LockManager.get_lock(bgp_speaker_id):
                peers = self._plugin.get_bgp_peers_by_bgp_speaker(
                    context, bgp_speaker_id)
                if bgp_peer_id not in [p['id'] for p in peers]:
                    continue
                bgp_bindings = nsxv_db.get_nsxv_bgp_speaker_bindings(
                    context.session, bgp_speaker_id)
                for binding in bgp_bindings:
                    try:
                        # Neighbours are identified by their ip address
                        self._nsxv.update_bgp_neighbours(binding['edge_id'],
                                                         [neighbour],
                                                         [neighbour])
                    except vcns_exc.VcnsApiException:
                        LOG.error("Failed to update BGP neighbor '%s' on "
                                  "edge '%s'", old_bgp_peer['peer_ip'],
                                  binding['edge_id'])

    def _validate_bgp_peer(self, context, bgp_speaker_id, new_peer_id):
        new_peer = self._plugin._get_bgp_peer(context, new_peer_id)
        peers = self._plugin._get_bgp_peers_by_bgp_speaker_binding(
            context, bgp_speaker_id)
        self._plugin._validate_peer_ips(bgp_speaker_id, peers, new_peer)

    def add_bgp_peer(self, context, bgp_speaker_id, bgp_peer_info):
        bgp_peer_id = self._plugin._get_id_for(bgp_peer_info, 'bgp_peer_id')
        bgp_peer_obj = self._plugin.get_bgp_peer(context,
                                                 bgp_peer_id)

        nbr = bgp_neighbour_from_peer(bgp_peer_obj)
        bgp_bindings = nsxv_db.get_nsxv_bgp_speaker_bindings(context.session,
                                                             bgp_speaker_id)
        self._validate_bgp_peer(context, bgp_speaker_id, bgp_peer_obj['id'])

        speaker = self._plugin.get_bgp_speaker(context, bgp_speaker_id)
        # list of tenant edge routers to be removed as bgp-neighbours to this
        # peer if it's associated with specific ESG.
        neighbours = []
        for binding in bgp_bindings:
            try:
                self._nsxv.add_bgp_neighbours(binding['edge_id'], [nbr])
            except vcns_exc.VcnsApiException:
                LOG.error("Failed to add BGP neighbour on '%s'",
                          binding['edge_id'])
            else:
                gw_nbr = gw_bgp_neighbour(binding['bgp_identifier'],
                                          speaker['local_as'],
                                          bgp_peer_obj['password'])
                neighbours.append(gw_nbr)
                LOG.debug("Succesfully added BGP neighbor '%s' on '%s'",
                          bgp_peer_obj['peer_ip'], binding['edge_id'])

        if bgp_peer_obj.get('esg_id'):
            edge_gw = bgp_peer_obj['esg_id']
            try:
                self._nsxv.add_bgp_neighbours(edge_gw, neighbours)
            except vcns_exc.VcnsApiException:
                with excutils.save_and_reraise_exception():
                    LOG.error("Failed to add BGP neighbour on GW Edge '%s'",
                              edge_gw)

    def remove_bgp_peer(self, context, bgp_speaker_id, bgp_peer_info):
        bgp_peer_id = bgp_peer_info['bgp_peer_id']
        bgp_peer_obj = self._plugin.get_bgp_peer(context, bgp_peer_id)
        nbr = bgp_neighbour_from_peer(bgp_peer_obj)
        bgp_bindings = nsxv_db.get_nsxv_bgp_speaker_bindings(
            context.session, bgp_speaker_id)
        speaker = self._plugin.get_bgp_speaker(context, bgp_speaker_id)
        # list of tenant edge routers to be removed as bgp-neighbours to this
        # peer if it's associated with specific ESG.
        neighbours = []
        for binding in bgp_bindings:
            try:
                self._nsxv.remove_bgp_neighbours(binding['edge_id'], [nbr])
            except vcns_exc.VcnsApiException:
                LOG.error("Failed to remove BGP neighbour on '%s'",
                          binding['edge_id'])
            else:
                gw_nbr = gw_bgp_neighbour(binding['bgp_identifier'],
                                          speaker['local_as'],
                                          bgp_peer_obj['password'])
                neighbours.append(gw_nbr)
                LOG.debug("Succesfully removed BGP neighbor '%s' on '%s'",
                          bgp_peer_obj['peer_ip'], binding['edge_id'])

        if bgp_peer_obj.get('esg_id'):
            edge_gw = bgp_peer_obj['esg_id']
            try:
                self._nsxv.remove_bgp_neighbours(edge_gw, neighbours)
            except vcns_exc.VcnsApiException:
                LOG.error("Failed to remove BGP neighbour on GW Edge '%s'",
                          edge_gw)

    def _validate_gateway_network(self, context, speaker_id, network_id):
        ext_net = self._core_plugin.get_network(context, network_id)

        if not ext_net.get(extnet_apidef.EXTERNAL):
            raise nsx_exc.NsxBgpNetworkNotExternal(net_id=network_id)
        if not ext_net['subnets']:
            raise nsx_exc.NsxBgpGatewayNetworkHasNoSubnets(net_id=network_id)

        # REVISIT(roeyc): Currently not allowing more than one bgp speaker per
        # gateway network.
        speakers_on_network = self._plugin._bgp_speakers_for_gateway_network(
            context, network_id)
        if speakers_on_network:
            raise bgp_ext.BgpSpeakerNetworkBindingError(
                network_id=network_id,
                bgp_speaker_id=speakers_on_network[0]['id'])

        subnet_id = ext_net['subnets'][0]
        ext_subnet = self._core_plugin.get_subnet(context, subnet_id)

        if ext_subnet.get('gateway_ip'):
            raise ext_esg_peer.ExternalSubnetHasGW(
                network_id=network_id, subnet_id=subnet_id)

        if not ext_net[address_scope.IPV4_ADDRESS_SCOPE]:
            raise nsx_exc.NsxBgpSpeakerUnableToAddGatewayNetwork(
                network_id=network_id, bgp_speaker_id=speaker_id)
        return True

    def add_gateway_network(self, context, bgp_speaker_id, network_info):
        gateway_network_id = network_info['network_id']

        if not self._validate_gateway_network(context, bgp_speaker_id,
                                              gateway_network_id):
            return

        edge_router_dict = self._get_dynamic_routing_edge_list(
            context, gateway_network_id, bgp_speaker_id)

        speaker = self._plugin.get_bgp_speaker(context, bgp_speaker_id)
        bgp_peers = self._plugin.get_bgp_peers_by_bgp_speaker(
            context, bgp_speaker_id)
        local_as = speaker['local_as']
        peers = []
        for edge_id, edge_router_config in edge_router_dict.items():
            router_ids = edge_router_config['no_snat_routers']
            advertise_static_routes = (
                edge_router_config['advertise_static_routes'])
            subnets = self._query_tenant_subnets(context, router_ids)
            # router_id here is in IP address format and is required for
            # the BGP configuration.
            bgp_identifier = edge_router_config['bgp_identifier']
            try:
                self._start_bgp_on_edge(context, edge_id, speaker,
                                        bgp_peers, bgp_identifier, subnets,
                                        advertise_static_routes)
            except vcns_exc.VcnsApiException:
                LOG.error("Failed to configure BGP speaker %s on edge '%s'.",
                          bgp_speaker_id, edge_id)
            else:
                peers.append(bgp_identifier)

        for edge_gw, password in [(peer['esg_id'], peer['password'])
                                  for peer in bgp_peers if peer.get('esg_id')]:
            neighbours = [gw_bgp_neighbour(bgp_id, local_as, password)
                          for bgp_id in peers]
            try:
                self._nsxv.add_bgp_neighbours(edge_gw, neighbours)
            except vcns_exc.VcnsApiException:
                LOG.error("Failed to add BGP neighbour on GW Edge '%s'",
                          edge_gw)

    def _start_bgp_on_edge(self, context, edge_id, speaker, bgp_peers,
                           bgp_identifier, subnets, advertise_static_routes):
        enabled_state = speaker['advertise_tenant_networks']
        local_as = speaker['local_as']
        prefixes, redis_rules = self._get_prefixes_and_redistribution_rules(
            subnets, advertise_static_routes)

        bgp_neighbours = [bgp_neighbour_from_peer(bgp_peer)
                          for bgp_peer in bgp_peers]
        try:
            self._nsxv.add_bgp_speaker_config(edge_id, bgp_identifier,
                                              local_as, enabled_state,
                                              bgp_neighbours, prefixes,
                                              redis_rules)
        except vcns_exc.VcnsApiException:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to configure BGP speaker '%s' on edge '%s'.",
                          speaker['id'], edge_id)
        else:
            nsxv_db.add_nsxv_bgp_speaker_binding(context.session, edge_id,
                                                 speaker['id'], bgp_identifier)

    def _stop_bgp_on_edges(self, context, bgp_bindings, speaker_id):
        peers_to_remove = []
        speaker = self._plugin.get_bgp_speaker(context, speaker_id)
        local_as = speaker['local_as']
        for bgp_binding in bgp_bindings:
            edge_id = bgp_binding['edge_id']
            try:
                self._nsxv.delete_bgp_speaker_config(edge_id)
            except vcns_exc.VcnsApiException:
                LOG.error("Failed to delete BGP speaker '%s' config on edge "
                          "'%s'.", speaker_id, edge_id)
            else:
                nsxv_db.delete_nsxv_bgp_speaker_binding(context.session,
                                                        edge_id)
                peers_to_remove.append(bgp_binding['bgp_identifier'])

        # We should also remove all bgp neighbours on gw-edges which
        # corresponds with tenant routers that are associated with this bgp
        # speaker.
        bgp_peers = self._plugin.get_bgp_peers_by_bgp_speaker(context,
                                                              speaker_id)
        gw_edges = [(peer['esg_id'], peer['password'])
                    for peer in bgp_peers if peer.get('esg_id')]
        for gw_edge, password in gw_edges:
            neighbours_to_remove = [gw_bgp_neighbour(bgp_identifier,
                                                     local_as,
                                                     password)
                                    for bgp_identifier in peers_to_remove]
            try:
                self._nsxv.remove_bgp_neighbours(gw_edge, neighbours_to_remove)
            except vcns_exc.VcnsApiException:
                LOG.error("Failed to remove BGP neighbour on GW edge '%s'.",
                          gw_edge)

    def remove_gateway_network(self, context, bgp_speaker_id, network_info):
        bgp_bindings = nsxv_db.get_nsxv_bgp_speaker_bindings(
            context.session, bgp_speaker_id)
        self._stop_bgp_on_edges(context, bgp_bindings, bgp_speaker_id)

    def _update_edge_bgp_identifier(self, context, bgp_binding, speaker,
                                    new_bgp_identifier):
        local_as = speaker['local_as']
        bgp_peers = self._plugin.get_bgp_peers_by_bgp_speaker(context,
                                                              speaker['id'])
        self._nsxv.update_router_id(bgp_binding['edge_id'], new_bgp_identifier)
        for gw_edge_id, password in [(peer['esg_id'], peer['password'])
                                     for peer in bgp_peers
                                     if peer.get('esg_id')]:
            nbr_to_remove = gw_bgp_neighbour(bgp_binding['bgp_identifier'],
                                             local_as, password)
            nbr_to_add = gw_bgp_neighbour(new_bgp_identifier, local_as,
                                          password)
            self._nsxv.update_bgp_neighbours(gw_edge_id,
                                             [nbr_to_add],
                                             [nbr_to_remove])

        with context.session.begin(subtransactions=True):
            bgp_binding['bgp_identifier'] = new_bgp_identifier

    def process_router_gw_port_update(self, context, speaker,
                                      router, updated_port):
        router_id = router['id']
        gw_fixed_ip = router.gw_port['fixed_ips'][0]['ip_address']

        edge_id, advertise_static_routes = (
            self._get_router_edge_info(context, router_id))
        if not edge_id:
            # shared router is not attached on any edge
            return

        bgp_binding = nsxv_db.get_nsxv_bgp_speaker_binding(
            context.session, edge_id)

        if bgp_binding:
            new_fixed_ip = updated_port['fixed_ips'][0]['ip_address']
            fixed_ip_updated = gw_fixed_ip != new_fixed_ip
            subnets = self._query_tenant_subnets(context, [router_id])
            prefixes, redis_rules = (
                self._get_prefixes_and_redistribution_rules(
                    subnets, advertise_static_routes))
            # Handle possible snat/no-nat update
            if router.enable_snat:
                self._nsxv.remove_bgp_redistribution_rules(edge_id, prefixes)
            else:
                self._nsxv.add_bgp_redistribution_rules(edge_id, prefixes,
                                                        redis_rules)
            if bgp_binding['bgp_identifier'] == gw_fixed_ip:
                if fixed_ip_updated:
                    self._update_edge_bgp_identifier(context,
                                                     bgp_binding,
                                                     speaker,
                                                     new_fixed_ip)

    def enable_bgp_on_router(self, context, speaker, router_id):
        local_as = speaker['local_as']
        edge_id, advertise_static_routes = (
            self._get_router_edge_info(context, router_id))
        if not edge_id:
            # shared router is not attached on any edge
            return
        router = self._core_plugin._get_router(context, router_id)
        subnets = self._query_tenant_subnets(context, [router_id])

        bgp_peers = self._plugin.get_bgp_peers_by_bgp_speaker(
            context, speaker['id'])
        bgp_binding = nsxv_db.get_nsxv_bgp_speaker_binding(
            context.session, edge_id)

        if bgp_binding and subnets:
            # Edge already configured with BGP (e.g - shared router edge),
            # Add the router attached subnets.
            if router.enable_snat:
                prefixes = [self.prefix_name(subnet['id'])
                            for subnet in subnets]
                self._nsxv.remove_bgp_redistribution_rules(edge_id, prefixes)
            else:
                prefixes, redis_rules = (
                    self._get_prefixes_and_redistribution_rules(
                        subnets, advertise_static_routes))
                self._nsxv.add_bgp_redistribution_rules(edge_id, prefixes,
                                                        redis_rules)
        elif not bgp_binding:
            if router.enable_snat:
                subnets = []
            bgp_identifier = router.gw_port['fixed_ips'][0]['ip_address']
            self._start_bgp_on_edge(context, edge_id, speaker, bgp_peers,
                                    bgp_identifier, subnets,
                                    advertise_static_routes)
            for gw_edge_id, password in [(peer['esg_id'], peer['password'])
                                         for peer in bgp_peers
                                         if peer.get('esg_id')]:
                nbr = gw_bgp_neighbour(bgp_identifier, local_as, password)
                self._nsxv.add_bgp_neighbours(gw_edge_id, [nbr])

    def disable_bgp_on_router(self, context, speaker, router_id, gw_ip,
                              edge_id=None):
        speaker = self._plugin.get_bgp_speaker(context, speaker['id'])
        current_edge_id, advertise_static_routes = (
            self._get_router_edge_info(context, router_id))
        edge_id = edge_id or current_edge_id

        if not edge_id:
            return

        bgp_binding = nsxv_db.get_nsxv_bgp_speaker_binding(context.session,
                                                           edge_id)
        if not bgp_binding:
            return

        # Need to ensure that we do not use the metadata IP's
        md_proxy = self._get_md_proxy_for_router(context, router_id)

        routers_ids = (
            self._core_plugin.edge_manager.get_routers_on_same_edge(
                context, router_id))
        routers_ids.remove(router_id)

        # We need to find out what other routers are hosted on the edges and
        # whether they have a gw addresses that could replace the current
        # bgp-identifier (if required).
        filters = {'device_owner': [n_const.DEVICE_OWNER_ROUTER_GW],
                   'device_id': routers_ids}
        edge_gw_ports = self._core_plugin.get_ports(context, filters=filters)
        alt_bgp_identifiers = [
            p['fixed_ips'][0]['ip_address'] for p in edge_gw_ports
            if (not md_proxy or
                not md_proxy.is_md_subnet(
                    p['fixed_ips'][0]['subnet_id']))]
        if alt_bgp_identifiers:
            # Shared router, only remove prefixes and redistribution
            # rules.
            subnets = self._query_tenant_subnets(context, [router_id])
            prefixes = [self.prefix_name(subnet['id'])
                        for subnet in subnets]
            self._nsxv.remove_bgp_redistribution_rules(edge_id, prefixes)
            if bgp_binding['bgp_identifier'] == gw_ip:
                self._update_edge_bgp_identifier(context, bgp_binding, speaker,
                                                 alt_bgp_identifiers[0])
        else:
            self._stop_bgp_on_edges(context, [bgp_binding], speaker['id'])

    def advertise_subnet(self, context, speaker_id, router_id, subnet):
        router = self._core_plugin._get_router(context, router_id)
        if router.enable_snat:
            # Do nothing, by default, only when advertisement is needed we add
            # a new redistribution rule
            return

        edge_id, advertise_static_routes = (
            self._get_router_edge_info(context, router_id))
        if not edge_id:
            # shared router is not attached on any edge
            return
        prefixes, redis_rules = self._get_prefixes_and_redistribution_rules(
            [subnet], advertise_static_routes)
        self._nsxv.add_bgp_redistribution_rules(edge_id, prefixes, redis_rules)

    def withdraw_subnet(self, context, speaker_id, router_id, subnet_id):
        router = self._core_plugin._get_router(context, router_id)
        if router.enable_snat:
            # Do nothing, by default, only when advertisement is needed we add
            # a new redistribution rule
            return

        edge_id, advertise_static_routes = (
            self._get_router_edge_info(context, router_id))
        prefix_name = self.prefix_name(subnet_id)
        self._nsxv.remove_bgp_redistribution_rules(edge_id, [prefix_name])



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""nsxv_security_group_logging

Revision ID: 2c87aedb206f
Revises: 4c45bcadccf9
Create Date: 2016-03-15 06:06:06.680092

"""

# revision identifiers, used by Alembic.
revision = '2c87aedb206f'
down_revision = '4c45bcadccf9'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('nsxv_security_group_section_mappings',
                  sa.Column('logging', sa.Boolean(), nullable=False))



# Copyright 2015 VMware Inc
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import optparse

import sqlalchemy as sa

from vmware_nsx.db import nsx_models
from vmware_nsxlib import v3
from vmware_nsxlib.v3 import config
from vmware_nsxlib.v3 import nsx_constants


class NeutronNsxDB(object):
    def __init__(self, db_connection):
        super(NeutronNsxDB, self).__init__()
        engine = sa.create_engine(db_connection)
        self.session = sa.orm.session.sessionmaker()(bind=engine)

    def query_all(self, column, model):
        return list(set([r[column] for r in self.session.query(model).all()]))

    def get_logical_ports(self):
        return self.query_all('nsx_port_id',
                              nsx_models.NeutronNsxPortMapping)

    def get_nsgroups(self):
        return self.query_all('nsx_id',
                              nsx_models.NeutronNsxSecurityGroupMapping)

    def get_firewall_sections(self):
        return self.query_all('nsx_id',
                              nsx_models.NeutronNsxFirewallSectionMapping)

    def get_logical_routers(self):
        return self.query_all('nsx_id',
                              nsx_models.NeutronNsxRouterMapping)

    def get_logical_switches(self):
        return self.query_all('nsx_id',
                              nsx_models.NeutronNsxNetworkMapping)

    def get_logical_dhcp_servers(self):
        return self.query_all('nsx_service_id',
                              nsx_models.NeutronNsxServiceBinding)

    def get_vpn_sessions(self):
        return self.query_all('session_id',
                              nsx_models.NsxVpnConnectionMapping)


class NSXClient(object):
    """Base NSX REST client"""
    API_VERSION = "v1"
    NULL_CURSOR_PREFIX = '0000'

    def __init__(self, host, username, password, db_connection):
        self.host = host
        self.username = username
        self.password = password
        self.neutron_db = (NeutronNsxDB(db_connection)
                           if db_connection else None)

        nsxlib_config = config.NsxLibConfig(
            username=self.username,
            password=self.password,
            nsx_api_managers=[self.host],
            # allow admin user to delete entities created
            # under openstack principal identity
            allow_overwrite_header=True)
        self.nsxlib = v3.NsxLib(nsxlib_config)

    def get_transport_zones(self):
        """
        Retrieve all transport zones
        """
        return self.nsxlib.transport_zone.list()['results']

    def get_logical_ports(self):
        """
        Retrieve all logical ports on NSX backend
        """
        return self.nsxlib.logical_port.list()['results']

    def get_os_logical_ports(self):
        """
        Retrieve all logical ports created from OpenStack
        """
        lports = self.get_os_resources(
            self.get_logical_ports())
        if self.neutron_db:
            db_lports = self.neutron_db.get_logical_ports()
            lports = [lp for lp in lports if lp['id'] in db_lports]
        return lports

    def update_logical_port_attachment(self, lports):
        """
        In order to delete logical ports, we need to detach
        the VIF attachment on the ports first.
        """
        for p in lports:
            try:
                self.nsxlib.logical_port.update(
                    p['id'], None, attachment_type=None)
            except Exception as e:
                print("ERROR: Failed to update lport %s: %s" % p['id'], e)

    def _remove_port_from_exclude_list(self, p):
        try:
            self.nsxlib.firewall_section.remove_member_from_fw_exclude_list(
                p['id'], None)
        except Exception:
            pass

    def _cleanup_logical_ports(self, lports):
        # logical port vif detachment
        self.update_logical_port_attachment(lports)
        for p in lports:
            # delete this port from the exclude list (if in it)
            self._remove_port_from_exclude_list(p)
            try:
                self.nsxlib.logical_port.delete(p['id'])
            except Exception as e:
                print("ERROR: Failed to delete logical port %s, error %s" %
                      (p['id'], e))
            else:
                print("Successfully deleted logical port %s" % p['id'])

    def cleanup_os_logical_ports(self):
        """
        Delete all logical ports created by OpenStack
        """
        os_lports = self.get_os_logical_ports()
        print("Number of OS Logical Ports to be deleted: %s" % len(os_lports))
        self._cleanup_logical_ports(os_lports)

    def get_os_resources(self, resources):
        """
        Get all logical resources created by OpenStack
        """
        os_resources = [r for r in resources if 'tags' in r
                        for tag in r['tags']
                        if 'os-api-version' in tag.values()]
        return os_resources

    def get_logical_switches(self):
        """
        Retrieve all logical switches on NSX backend
        """
        return self.nsxlib.logical_switch.list()['results']

    def get_os_logical_switches(self):
        """
        Retrieve all logical switches created from OpenStack
        """
        lswitches = self.get_os_resources(
            self.get_logical_switches())

        if self.neutron_db:
            db_lswitches = self.neutron_db.get_logical_switches()
            lswitches = [ls for ls in lswitches
                         if ls['id'] in db_lswitches]
        return lswitches

    def get_lswitch_ports(self, ls_id):
        """
        Return all the logical ports that belong to this lswitch
        """
        lports = self.get_logical_ports()
        return [p for p in lports if p['logical_switch_id'] == ls_id]

    def cleanup_os_logical_switches(self):
        """
        Delete all logical switches created from OpenStack
        """
        lswitches = self.get_os_logical_switches()
        print("Number of OS Logical Switches to be deleted: %s" %
              len(lswitches))
        for ls in lswitches:
            # Check if there are still ports on switch and blow them away
            # An example here is a metadata proxy port (this is not stored
            # in the DB so we are unable to delete it when reading ports
            # from the DB)
            lports = self.get_lswitch_ports(ls['id'])
            if lports:
                print("Number of orphan OS Logical Ports to be "
                      "deleted: %s" % len(lports))
                self._cleanup_logical_ports(lports)

            try:
                self.nsxlib.logical_switch.delete(ls['id'])
            except Exception as e:
                print("ERROR: Failed to delete logical switch %s-%s, "
                      "error %s" % (ls['display_name'], ls['id'], e))
            else:
                print("Successfully deleted logical switch %s-%s" %
                      (ls['display_name'], ls['id']))

    def get_firewall_sections(self):
        """
        Retrieve all firewall sections
        """
        return self.nsxlib.firewall_section.list()

    def get_os_firewall_sections(self):
        """
        Retrieve all firewall sections created from OpenStack
        """
        fw_sections = self.get_os_resources(
            self.get_firewall_sections())
        if self.neutron_db:
            db_sections = self.neutron_db.get_firewall_sections()
            fw_sections = [fws for fws in fw_sections
                           if fws['id'] in db_sections]
        return fw_sections

    def cleanup_os_firewall_sections(self):
        """
        Cleanup all firewall sections created from OpenStack
        """
        fw_sections = self.get_os_firewall_sections()
        print("Number of OS Firewall Sections to be deleted: %s" %
              len(fw_sections))
        for fw in fw_sections:
            try:
                self.nsxlib.firewall_section.delete(fw['id'])
            except Exception as e:
                print("Failed to delete firewall section %s: %s" %
                      (fw['display_name'], e))
            else:
                print("Successfully deleted firewall section %s" %
                      fw['display_name'])

    def get_ns_groups(self):
        """
        Retrieve all NSGroups on NSX backend
        """
        backend_groups = self.nsxlib.ns_group.list()
        ns_groups = self.get_os_resources(backend_groups)
        if self.neutron_db:
            db_nsgroups = self.neutron_db.get_nsgroups()
            ns_groups = [nsg for nsg in ns_groups
                         if nsg['id'] in db_nsgroups]
        return ns_groups

    def cleanup_os_ns_groups(self):
        """
        Cleanup all NSGroups created from OpenStack plugin
        """
        ns_groups = self.get_ns_groups()
        print("Number of OS NSGroups to be deleted: %s" % len(ns_groups))
        for nsg in ns_groups:
            try:
                self.nsxlib.ns_group.delete(nsg['id'])
            except Exception as e:
                print("Failed to delete NSGroup: %s: %s" %
                      (nsg['display_name'], e))
            else:
                print("Successfully deleted NSGroup: %s" %
                      nsg['display_name'])

    def get_switching_profiles(self):
        """
        Retrieve all Switching Profiles on NSX backend
        """
        return self.nsxlib.switching_profile.list()['results']

    def get_os_switching_profiles(self):
        """
        Retrieve all Switching Profiles created from OpenStack
        """
        sw_profiles = self.get_os_resources(
            self.get_switching_profiles())
        if self.neutron_db:
            sw_profiles = []
        return sw_profiles

    def cleanup_os_switching_profiles(self):
        """
        Cleanup all Switching Profiles created from OpenStack plugin
        """
        sw_profiles = self.get_os_switching_profiles()
        print("Number of OS SwitchingProfiles to be deleted: %s" %
              len(sw_profiles))
        for swp in sw_profiles:
            try:
                self.nsxlib.switching_profile.delete(swp['id'])
            except Exception as e:
                print("Failed to delete Switching Profile: %s: %s" %
                      (swp['display_name'], e))
            else:
                print("Successfully deleted Switching Profile: %s" %
                      swp['display_name'])

    def get_logical_routers(self, tier=None):
        """
        Retrieve all the logical routers based on router type. If tier
        is None, it will return all logical routers.
        """
        lrouters = self.nsxlib.logical_router.list(
            router_type=tier)['results']

        if self.neutron_db:
            db_routers = self.neutron_db.get_logical_routers()
            lrouters = [lr for lr in lrouters
                        if lr['id'] in db_routers]
        return lrouters

    def get_os_logical_routers(self):
        """
        Retrieve all logical routers created from Neutron NSXv3 plugin
        """
        lrouters = self.get_logical_routers()
        return self.get_os_resources(lrouters)

    def get_logical_router_ports(self, lrouter):
        """
        Get all logical ports attached to lrouter
        """
        return self.nsxlib.logical_router_port.get_by_router_id(lrouter['id'])

    def get_os_logical_router_ports(self, lrouter):
        """
        Retrieve all logical router ports created from Neutron NSXv3 plugin
        """
        lports = self.get_logical_router_ports(lrouter)
        return self.get_os_resources(lports)

    def cleanup_logical_router_ports(self, lrouter):
        """
        Cleanup all logical ports on a logical router
        """
        lports = self.get_os_logical_router_ports(lrouter)
        for lp in lports:
            try:
                self.nsxlib.logical_router_port.delete(lp['id'])
            except Exception as e:
                print("Failed to delete logical router port %s-%s, "
                      "and response is %s" %
                      (lp['display_name'], lp['id'], e))
            else:
                print("Successfully deleted logical router port %s-%s" %
                      (lp['display_name'], lp['id']))

    def cleanup_os_logical_routers(self):
        """
        Delete all logical routers created from OpenStack
        To delete a logical router, we need to delete all logical
        ports on the router first.
        """
        lrouters = self.get_os_logical_routers()
        print("Number of OS Logical Routers to be deleted: %s" %
              len(lrouters))
        for lr in lrouters:
            self.cleanup_logical_router_ports(lr)
            self.cleanup_logical_router_vpn_sess(lr)
            try:
                self.nsxlib.logical_router.delete(lr['id'])
            except Exception as e:
                print("ERROR: Failed to delete logical router %s-%s, "
                      "error %s" % (lr['display_name'], lr['id'], e))
            else:
                print("Successfully deleted logical router %s-%s" %
                      (lr['display_name'], lr['id']))

    def cleanup_os_tier0_logical_ports(self):
        """
        Delete all TIER0 logical router ports created from OpenStack
        """
        tier0_routers = self.get_logical_routers(tier='TIER0')
        for lr in tier0_routers:
            self.cleanup_logical_router_ports(lr)

    def get_logical_dhcp_servers(self):
        """
        Retrieve all logical DHCP servers on NSX backend
        """
        return self.nsxlib.dhcp_server.list()['results']

    def get_os_logical_dhcp_servers(self):
        """
        Retrieve all logical DHCP servers created from OpenStack
        """
        dhcp_servers = self.get_os_resources(
            self.get_logical_dhcp_servers())

        if self.neutron_db:
            db_dhcp_servers = self.neutron_db.get_logical_dhcp_servers()
            dhcp_servers = [srv for srv in dhcp_servers
                            if srv['id'] in db_dhcp_servers]
        return dhcp_servers

    def cleanup_os_logical_dhcp_servers(self):
        """
        Cleanup all logical DHCP servers created from OpenStack plugin
        """
        dhcp_servers = self.get_os_logical_dhcp_servers()
        print("Number of OS Logical DHCP Servers to be deleted: %s" %
              len(dhcp_servers))
        for server in dhcp_servers:
            try:
                self.nsxlib.dhcp_server.delete(server['id'])
            except Exception as e:
                print("ERROR: Failed to delete logical DHCP server %s, "
                      "error %s" % (server['display_name'], e))
            else:
                print("Successfully deleted logical DHCP server %s" %
                      server['display_name'])

    def get_os_vpn_sessions(self):
        """
        Retrieve all nsx vpn sessions from nsx and OpenStack
        """
        sessions = self.get_os_resources(
            self.nsxlib.vpn_ipsec.session.list()['results'])

        if self.neutron_db:
            db_sessions = self.neutron_db.get_vpn_sessions()
            sessions = [sess for sess in sessions
                        if sess['id'] in db_sessions]
        return sessions

    def cleanup_vpnaas_objects(self):
        """
        Cleanup vpn/ipsec nsx objects
        """
        if not self.nsxlib.feature_supported(nsx_constants.FEATURE_IPSEC_VPN):
            # no vpn support
            return

        # sessions: leftover sessions prevent us from configuring new similar
        # sessions so it is important to delete them
        sessions = self.get_os_vpn_sessions()
        for session in sessions:
            try:
                self.nsxlib.vpn_ipsec.session.delete(session['id'])
            except Exception as e:
                print("ERROR: Failed to delete vpn ipsec session %s, "
                      "error %s" % (session['id'], e))
            else:
                print("Successfully deleted vpn ipsec session %s" %
                      session['id'])

    def cleanup_logical_router_vpn_sess(self, lr):
        """
        Cleanup the vpn local session of the logical router
        """
        if not self.nsxlib.feature_supported(nsx_constants.FEATURE_IPSEC_VPN):
            # no vpn support
            return

        # find the router neutron id in its tags
        neutron_id = None
        for tag in lr['tags']:
            if tag.get('scope') == 'os-neutron-router-id':
                neutron_id = tag.get('tag')
                break

        if not neutron_id:
            return

        tags = [{'scope': 'os-neutron-router-id', 'tag': neutron_id}]
        ep_list = self.nsxlib.search_by_tags(
            tags=tags,
            resource_type=self.nsxlib.vpn_ipsec.local_endpoint.resource_type)
        if ep_list['results']:
            id = ep_list['results'][0]['id']
            try:
                self.nsxlib.vpn_ipsec.local_endpoint.delete(id)
            except Exception as e:
                print("ERROR: Failed to delete vpn ipsec local endpoint %s, "
                      "error %s" % (id, e))
            else:
                print("Successfully deleted vpn ipsec local endpoint %s" % id)

    def cleanup_all(self):
        """
        Cleanup steps:
            1. Cleanup firewall sections
            2. Cleanup NSGroups
            3. Cleanup logical router ports
            4. Cleanup logical routers
            5. Cleanup logical switch ports
            6. Cleanup logical switches
            7. Cleanup switching profiles
        """
        self.cleanup_os_firewall_sections()
        self.cleanup_os_ns_groups()
        self.cleanup_vpnaas_objects()
        self.cleanup_os_logical_routers()
        self.cleanup_os_tier0_logical_ports()
        self.cleanup_os_logical_ports()
        self.cleanup_os_logical_switches()
        self.cleanup_os_logical_dhcp_servers()
        self.cleanup_os_switching_profiles()


if __name__ == "__main__":

    parser = optparse.OptionParser()
    parser.add_option("--mgr-ip", dest="mgr_ip", help="NSX Manager IP address")
    parser.add_option("-u", "--username", default="admin", dest="username",
                      help="NSX Manager username")
    parser.add_option("-p", "--password", default="default", dest="password",
                      help="NSX Manager password")
    parser.add_option("--db-connection", default="", dest="db_connection",
                      help=("When set, cleaning only backend resources that "
                            "have db record."))
    (options, args) = parser.parse_args()

    # Get NSX REST client
    nsx_client = NSXClient(options.mgr_ip, options.username,
                           options.password, options.db_connection)
    # Clean all objects created by OpenStack
    nsx_client.cleanup_all()



# Copyright 2016 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""NSXv add availability zone to the router bindings table instead of
the resource pool column

Revision ID: 86a55205337c
Revises: 7e46906f8997
Create Date: 2016-07-12 09:18:44.450116
"""

# revision identifiers, used by Alembic.
revision = '86a55205337c'
down_revision = '7e46906f8997'

from alembic import op
import sqlalchemy as sa

from vmware_nsx.common import config  # noqa


def upgrade():
    op.alter_column('nsxv_router_bindings', 'resource_pool',
                    new_column_name='availability_zone',
                    existing_type=sa.String(36),
                    existing_nullable=True,
                    existing_server_default='default')



# Copyright 2013 VMware, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_config import cfg
from oslo_log import log as logging

from vmware_nsx.plugins.nsx_v.vshield import edge_appliance_driver
from vmware_nsx.plugins.nsx_v.vshield import edge_dynamic_routing_driver
from vmware_nsx.plugins.nsx_v.vshield import edge_firewall_driver
from vmware_nsx.plugins.nsx_v.vshield.tasks import tasks
from vmware_nsx.plugins.nsx_v.vshield import vcns
from vmware_nsx.services.lbaas.nsx_v.v2 import (
    edge_loadbalancer_driver_v2 as lbaas_v2)

LOG = logging.getLogger(__name__)


class VcnsDriver(edge_appliance_driver.EdgeApplianceDriver,
                 lbaas_v2.EdgeLoadbalancerDriverV2,
                 edge_firewall_driver.EdgeFirewallDriver,
                 edge_dynamic_routing_driver.EdgeDynamicRoutingDriver):

    def __init__(self, callbacks):
        super(VcnsDriver, self).__init__()

        self.callbacks = callbacks
        self.vcns_uri = cfg.CONF.nsxv.manager_uri
        self.vcns_user = cfg.CONF.nsxv.user
        self.vcns_passwd = cfg.CONF.nsxv.password
        self.ca_file = cfg.CONF.nsxv.ca_file
        self.insecure = cfg.CONF.nsxv.insecure
        self.deployment_container_id = cfg.CONF.nsxv.deployment_container_id
        self._pid = None
        self._task_manager = None
        self.vcns = vcns.Vcns(self.vcns_uri, self.vcns_user, self.vcns_passwd,
                              self.ca_file, self.insecure)

    @property
    def task_manager(self):
        if (self._task_manager is None or
                self._pid != os.getpid()):
            LOG.debug("Creating task manager")
            self._pid = os.getpid()
            interval = cfg.CONF.nsxv.task_status_check_interval
            self._task_manager = tasks.TaskManager(interval)
            LOG.debug("Starting task manager")
            self._task_manager.start()
        return self._task_manager



# Copyright 2013 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy.orm import exc as sa_orm_exc

from neutron.db import _model_query as model_query
from neutron.db import _utils as db_utils
from neutron.db import api as db_api
from neutron.plugins.common import utils
from neutron_lib import constants
from neutron_lib import exceptions
from oslo_log import log as logging
from oslo_utils import uuidutils
import six

from vmware_nsx._i18n import _
from vmware_nsx.db import nsx_models
from vmware_nsx.extensions import networkgw

LOG = logging.getLogger(__name__)
DEVICE_OWNER_NET_GW_INTF = 'network:gateway-interface'
NETWORK_ID = 'network_id'
SEGMENTATION_TYPE = 'segmentation_type'
SEGMENTATION_ID = 'segmentation_id'
ALLOWED_CONNECTION_ATTRIBUTES = set((NETWORK_ID,
                                     SEGMENTATION_TYPE,
                                     SEGMENTATION_ID))
# Constants for gateway device operational status
STATUS_UNKNOWN = "UNKNOWN"
STATUS_ERROR = "ERROR"
STATUS_ACTIVE = "ACTIVE"
STATUS_DOWN = "DOWN"


class GatewayInUse(exceptions.InUse):
    message = _("Network Gateway '%(gateway_id)s' still has active mappings "
                "with one or more neutron networks.")


class GatewayNotFound(exceptions.NotFound):
    message = _("Network Gateway %(gateway_id)s could not be found")


class GatewayDeviceInUse(exceptions.InUse):
    message = _("Network Gateway Device '%(device_id)s' is still used by "
                "one or more network gateways.")


class GatewayDeviceNotFound(exceptions.NotFound):
    message = _("Network Gateway Device %(device_id)s could not be found.")


class GatewayDevicesNotFound(exceptions.NotFound):
    message = _("One or more Network Gateway Devices could not be found: "
                "%(device_ids)s.")


class NetworkGatewayPortInUse(exceptions.InUse):
    message = _("Port '%(port_id)s' is owned by '%(device_owner)s' and "
                "therefore cannot be deleted directly via the port API.")


class GatewayConnectionInUse(exceptions.InUse):
    message = _("The specified mapping '%(mapping)s' is already in use on "
                "network gateway '%(gateway_id)s'.")


class MultipleGatewayConnections(exceptions.Conflict):
    message = _("Multiple network connections found on '%(gateway_id)s' "
                "with provided criteria.")


class GatewayConnectionNotFound(exceptions.NotFound):
    message = _("The connection %(network_mapping_info)s was not found on the "
                "network gateway '%(network_gateway_id)s'")


class NetworkGatewayUnchangeable(exceptions.InUse):
    message = _("The network gateway %(gateway_id)s "
                "cannot be updated or deleted")


class NetworkGatewayMixin(networkgw.NetworkGatewayPluginBase):

    gateway_resource = networkgw.GATEWAY_RESOURCE_NAME
    device_resource = networkgw.DEVICE_RESOURCE_NAME

    def _get_network_gateway(self, context, gw_id):
        try:
            gw = model_query.get_by_id(context, nsx_models.NetworkGateway,
                                       gw_id)
        except sa_orm_exc.NoResultFound:
            raise GatewayNotFound(gateway_id=gw_id)
        return gw

    def _make_gw_connection_dict(self, gw_conn):
        return {'port_id': gw_conn['port_id'],
                'segmentation_type': gw_conn['segmentation_type'],
                'segmentation_id': gw_conn['segmentation_id']}

    def _make_network_gateway_dict(self, network_gateway, fields=None):
        device_list = []
        for d in network_gateway['devices']:
            device_list.append({'id': d['id'],
                                'interface_name': d['interface_name']})
        res = {'id': network_gateway['id'],
               'name': network_gateway['name'],
               'default': network_gateway['default'],
               'devices': device_list,
               'tenant_id': network_gateway['tenant_id']}
        # Query gateway connections only if needed
        if not fields or 'ports' in fields:
            res['ports'] = [self._make_gw_connection_dict(conn)
                            for conn in network_gateway.network_connections]
        return db_utils.resource_fields(res, fields)

    def _set_mapping_info_defaults(self, mapping_info):
        if not mapping_info.get('segmentation_type'):
            mapping_info['segmentation_type'] = 'flat'
        if not mapping_info.get('segmentation_id'):
            mapping_info['segmentation_id'] = 0

    def _validate_network_mapping_info(self, network_mapping_info):
        self._set_mapping_info_defaults(network_mapping_info)
        network_id = network_mapping_info.get(NETWORK_ID)
        if not network_id:
            raise exceptions.InvalidInput(
                error_message=_("A network identifier must be specified "
                                "when connecting a network to a network "
                                "gateway. Unable to complete operation"))
        connection_attrs = set(network_mapping_info.keys())
        if not connection_attrs.issubset(ALLOWED_CONNECTION_ATTRIBUTES):
            raise exceptions.InvalidInput(
                error_message=(_("Invalid keys found among the ones provided "
                                 "in request body: %(connection_attrs)s."),
                               connection_attrs))
        seg_type = network_mapping_info.get(SEGMENTATION_TYPE)
        seg_id = network_mapping_info.get(SEGMENTATION_ID)
        # It is important to validate that the segmentation ID is actually an
        # integer value
        try:
            seg_id = int(seg_id)
        except ValueError:
            msg = _("An invalid segmentation ID was specified. The "
                    "segmentation ID must be a positive integer number")
            raise exceptions.InvalidInput(error_message=msg)
        # The NSX plugin accepts 0 as a valid vlan tag
        seg_id_valid = seg_id == 0 or utils.is_valid_vlan_tag(seg_id)
        if seg_type.lower() == 'flat' and seg_id:
            msg = _("Cannot specify a segmentation id when "
                    "the segmentation type is flat")
            raise exceptions.InvalidInput(error_message=msg)
        elif (seg_type.lower() == 'vlan' and not seg_id_valid):
            msg = _("Invalid segmentation id (%s) for "
                    "vlan segmentation type") % seg_id
            raise exceptions.InvalidInput(error_message=msg)
        return network_id

    def _retrieve_gateway_connections(self, context, gateway_id,
                                      mapping_info=None, only_one=False):
        mapping_info = mapping_info or {}
        filters = {'network_gateway_id': [gateway_id]}
        for k, v in six.iteritems(mapping_info):
            if v and k != NETWORK_ID:
                filters[k] = [v]
        query = model_query.get_collection_query(context,
                                                 nsx_models.NetworkConnection,
                                                 filters)
        return query.one() if only_one else query.all()

    def _unset_default_network_gateways(self, context):
        with db_api.context_manager.writer.using(context):
            context.session.query(nsx_models.NetworkGateway).update(
                {nsx_models.NetworkGateway.default: False})

    def _set_default_network_gateway(self, context, gw_id):
        with db_api.context_manager.writer.using(context):
            gw = (context.session.query(nsx_models.NetworkGateway).
                  filter_by(id=gw_id).one())
            gw['default'] = True

    def prevent_network_gateway_port_deletion(self, context, port):
        """Pre-deletion check.

        Ensures a port will not be deleted if is being used by a network
        gateway. In that case an exception will be raised.
        """
        if port['device_owner'] == DEVICE_OWNER_NET_GW_INTF:
            raise NetworkGatewayPortInUse(port_id=port['id'],
                                          device_owner=port['device_owner'])

    def _validate_device_list(self, context, tenant_id, gateway_data):
        device_query = self._query_gateway_devices(
            context, filters={'id': [device['id']
                                     for device in gateway_data['devices']]})
        retrieved_device_ids = set()
        for device in device_query:
            retrieved_device_ids.add(device['id'])
            if device['tenant_id'] != tenant_id:
                raise GatewayDeviceNotFound(device_id=device['id'])
        missing_device_ids = (
            set(device['id'] for device in gateway_data['devices']) -
            retrieved_device_ids)
        if missing_device_ids:
            raise GatewayDevicesNotFound(
                device_ids=",".join(missing_device_ids))

    def create_network_gateway(self, context, network_gateway,
            validate_device_list=True):
        gw_data = network_gateway[self.gateway_resource]
        tenant_id = gw_data['tenant_id']
        with db_api.context_manager.writer.using(context):
            gw_db = nsx_models.NetworkGateway(
                id=gw_data.get('id', uuidutils.generate_uuid()),
                tenant_id=tenant_id,
                name=gw_data.get('name'))
            # Device list is guaranteed to be a valid list, but some devices
            # might still either not exist or belong to a different tenant
            if validate_device_list:
                self._validate_device_list(context, tenant_id, gw_data)
            gw_db.devices.extend(
                [nsx_models.NetworkGatewayDeviceReference(**device)
                 for device in gw_data['devices']])
            context.session.add(gw_db)
            LOG.debug("Created network gateway with id:%s", gw_db['id'])
            return self._make_network_gateway_dict(gw_db)

    def update_network_gateway(self, context, id, network_gateway):
        gw_data = network_gateway[self.gateway_resource]
        with db_api.context_manager.writer.using(context):
            gw_db = self._get_network_gateway(context, id)
            if gw_db.default:
                raise NetworkGatewayUnchangeable(gateway_id=id)
            # Ensure there is something to update before doing it
            if any([gw_db[k] != gw_data[k] for k in gw_data]):
                gw_db.update(gw_data)
            LOG.debug("Updated network gateway with id:%s", id)
            return self._make_network_gateway_dict(gw_db)

    def get_network_gateway(self, context, id, fields=None):
        gw_db = self._get_network_gateway(context, id)
        return self._make_network_gateway_dict(gw_db, fields)

    def delete_network_gateway(self, context, id):
        with db_api.context_manager.writer.using(context):
            gw_db = self._get_network_gateway(context, id)
            if gw_db.network_connections:
                raise GatewayInUse(gateway_id=id)
            if gw_db.default:
                raise NetworkGatewayUnchangeable(gateway_id=id)
            context.session.delete(gw_db)
        LOG.debug("Network gateway '%s' was destroyed.", id)

    def get_network_gateways(self, context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        marker_obj = db_utils.get_marker_obj(self,
            context, 'network_gateway', limit, marker)
        return model_query.get_collection(context, nsx_models.NetworkGateway,
                                          self._make_network_gateway_dict,
                                          filters=filters, fields=fields,
                                          sorts=sorts, limit=limit,
                                          marker_obj=marker_obj,
                                          page_reverse=page_reverse)

    def connect_network(self, context, network_gateway_id,
                        network_mapping_info):
        network_id = self._validate_network_mapping_info(network_mapping_info)
        LOG.debug("Connecting network '%(network_id)s' to gateway "
                  "'%(network_gateway_id)s'",
                  {'network_id': network_id,
                   'network_gateway_id': network_gateway_id})
        with db_api.context_manager.writer.using(context):
            gw_db = self._get_network_gateway(context, network_gateway_id)
            tenant_id = gw_db['tenant_id']
            if context.is_admin and not tenant_id:
                tenant_id = context.tenant_id
            # TODO(salvatore-orlando): Leverage unique constraint instead
            # of performing another query!
            if self._retrieve_gateway_connections(context,
                                                  network_gateway_id,
                                                  network_mapping_info):
                raise GatewayConnectionInUse(mapping=network_mapping_info,
                                             gateway_id=network_gateway_id)
            # TODO(salvatore-orlando): Creating a port will give it an IP,
            # but we actually do not need any. Instead of wasting an IP we
            # should have a way to say a port shall not be associated with
            # any subnet
            try:
                # We pass the segmentation type and id too - the plugin
                # might find them useful as the network connection object
                # does not exist yet.
                # NOTE: they're not extended attributes, rather extra data
                # passed in the port structure to the plugin
                # TODO(salvatore-orlando): Verify optimal solution for
                # ownership of the gateway port
                port = self.create_port(context, {
                    'port':
                    {'tenant_id': tenant_id,
                     'network_id': network_id,
                     'mac_address': constants.ATTR_NOT_SPECIFIED,
                     'admin_state_up': True,
                     'fixed_ips': [],
                     'device_id': network_gateway_id,
                     'device_owner': DEVICE_OWNER_NET_GW_INTF,
                     'name': '',
                     'gw:segmentation_type':
                     network_mapping_info.get('segmentation_type'),
                     'gw:segmentation_id':
                     network_mapping_info.get('segmentation_id')}})
            except exceptions.NetworkNotFound:
                err_msg = (_("Requested network '%(network_id)s' not found."
                             "Unable to create network connection on "
                             "gateway '%(network_gateway_id)s") %
                           {'network_id': network_id,
                            'network_gateway_id': network_gateway_id})
                LOG.error(err_msg)
                raise exceptions.InvalidInput(error_message=err_msg)
            port_id = port['id']
            LOG.debug("Gateway port for '%(network_gateway_id)s' "
                      "created on network '%(network_id)s':%(port_id)s",
                      {'network_gateway_id': network_gateway_id,
                       'network_id': network_id,
                       'port_id': port_id})
            # Create NetworkConnection record
            network_mapping_info['port_id'] = port_id
            network_mapping_info['tenant_id'] = tenant_id
            gw_db.network_connections.append(
                nsx_models.NetworkConnection(**network_mapping_info))
            port_id = port['id']
            # now deallocate and recycle ip from the port
            for fixed_ip in port.get('fixed_ips', []):
                self._delete_ip_allocation(context, network_id,
                                           fixed_ip['subnet_id'],
                                           fixed_ip['ip_address'])
            LOG.debug("Ensured no Ip addresses are configured on port %s",
                      port_id)
            return {'connection_info':
                    {'network_gateway_id': network_gateway_id,
                     'network_id': network_id,
                     'port_id': port_id}}

    def disconnect_network(self, context, network_gateway_id,
                           network_mapping_info):
        network_id = self._validate_network_mapping_info(network_mapping_info)
        LOG.debug("Disconnecting network '%(network_id)s' from gateway "
                  "'%(network_gateway_id)s'",
                  {'network_id': network_id,
                   'network_gateway_id': network_gateway_id})
        with db_api.context_manager.writer.using(context):
            # Uniquely identify connection, otherwise raise
            try:
                net_connection = self._retrieve_gateway_connections(
                    context, network_gateway_id,
                    network_mapping_info, only_one=True)
            except sa_orm_exc.NoResultFound:
                raise GatewayConnectionNotFound(
                    network_mapping_info=network_mapping_info,
                    network_gateway_id=network_gateway_id)
            except sa_orm_exc.MultipleResultsFound:
                raise MultipleGatewayConnections(
                    gateway_id=network_gateway_id)
            # Remove gateway port from network
            # FIXME(salvatore-orlando): Ensure state of port in NSX is
            # consistent with outcome of transaction
            self.delete_port(context, net_connection['port_id'],
                             nw_gw_port_check=False)
            # Remove NetworkConnection record
            context.session.delete(net_connection)

    def _make_gateway_device_dict(self, gateway_device, fields=None,
                                  include_nsx_id=False):
        res = {'id': gateway_device['id'],
               'name': gateway_device['name'],
               'status': gateway_device['status'],
               'connector_type': gateway_device['connector_type'],
               'connector_ip': gateway_device['connector_ip'],
               'tenant_id': gateway_device['tenant_id']}
        if include_nsx_id:
            # Return the NSX mapping as well. This attribute will not be
            # returned in the API response anyway. Ensure it will not be
            # filtered out in field selection.
            if fields:
                fields.append('nsx_id')
            res['nsx_id'] = gateway_device['nsx_id']
        return db_utils.resource_fields(res, fields)

    def _get_gateway_device(self, context, device_id):
        try:
            return model_query.get_by_id(context,
                                         nsx_models.NetworkGatewayDevice,
                                         device_id)
        except sa_orm_exc.NoResultFound:
            raise GatewayDeviceNotFound(device_id=device_id)

    def _is_device_in_use(self, context, device_id):
        query = model_query.get_collection_query(
            context, nsx_models.NetworkGatewayDeviceReference,
            {'id': [device_id]})
        return query.first()

    def get_gateway_device(self, context, device_id, fields=None,
                           include_nsx_id=False):
        return self._make_gateway_device_dict(
            self._get_gateway_device(context, device_id),
            fields, include_nsx_id)

    def _query_gateway_devices(self, context,
                               filters=None, sorts=None,
                               limit=None, marker=None,
                               page_reverse=None):
        marker_obj = db_utils.get_marker_obj(self,
            context, 'gateway_device', limit, marker)
        return self._get_collection_query(context,
                                          nsx_models.NetworkGatewayDevice,
                                          filters=filters,
                                          sorts=sorts,
                                          limit=limit,
                                          marker_obj=marker_obj,
                                          page_reverse=page_reverse)

    def get_gateway_devices(self, context, filters=None, fields=None,
                            sorts=None, limit=None, marker=None,
                            page_reverse=False, include_nsx_id=False):
        query = self._query_gateway_devices(context, filters, sorts, limit,
                                            marker, page_reverse)
        return [self._make_gateway_device_dict(row, fields, include_nsx_id)
                for row in query]

    def create_gateway_device(self, context, gateway_device,
                              initial_status=STATUS_UNKNOWN):
        device_data = gateway_device[self.device_resource]
        tenant_id = device_data['tenant_id']
        with db_api.context_manager.writer.using(context):
            device_db = nsx_models.NetworkGatewayDevice(
                id=device_data.get('id', uuidutils.generate_uuid()),
                tenant_id=tenant_id,
                name=device_data.get('name'),
                connector_type=device_data['connector_type'],
                connector_ip=device_data['connector_ip'],
                status=initial_status)
            context.session.add(device_db)
        LOG.debug("Created network gateway device: %s", device_db['id'])
        return self._make_gateway_device_dict(device_db)

    def update_gateway_device(self, context, gateway_device_id,
                              gateway_device, include_nsx_id=False):
        device_data = gateway_device[self.device_resource]
        with db_api.context_manager.writer.using(context):
            device_db = self._get_gateway_device(context, gateway_device_id)
            # Ensure there is something to update before doing it
            if any([device_db[k] != device_data[k] for k in device_data]):
                device_db.update(device_data)
        LOG.debug("Updated network gateway device: %s",
                  gateway_device_id)
        return self._make_gateway_device_dict(
            device_db, include_nsx_id=include_nsx_id)

    def delete_gateway_device(self, context, device_id):
        with db_api.context_manager.writer.using(context):
            # A gateway device should not be deleted
            # if it is used in any network gateway service
            if self._is_device_in_use(context, device_id):
                raise GatewayDeviceInUse(device_id=device_id)
            device_db = self._get_gateway_device(context, device_id)
            context.session.delete(device_db)
        LOG.debug("Deleted network gateway device: %s.", device_id)



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.services.fwaas.common import fwaas_callbacks_v1 as com_clbcks
from vmware_nsx.services.fwaas.nsx_tv import edge_fwaas_driver_v1 as tv_driver

LOG = logging.getLogger(__name__)


class Nsxv3FwaasCallbacksV1(com_clbcks.NsxFwaasCallbacks):
    """NSX-V3 RPC callbacks for Firewall As A Service - V1."""

    def __init__(self):
        super(Nsxv3FwaasCallbacksV1, self).__init__()
        # update the fwaas driver in case of TV plugin
        if self.fwaas_enabled:
            if self.fwaas_driver.driver_name == tv_driver.FWAAS_DRIVER_NAME:
                self.internal_driver = self.fwaas_driver.get_T_driver()
            else:
                self.internal_driver = self.fwaas_driver

    @property
    def plugin_type(self):
        return projectpluginmap.NsxPlugins.NSX_T

    def should_apply_firewall_to_router(self, context, router_id):
        """Return True if the FWaaS rules should be added to this router."""
        if not super(Nsxv3FwaasCallbacksV1,
                     self).should_apply_firewall_to_router(context,
                                                           router_id):
            return False

        # get all the relevant router info
        ctx_elevated = context.elevated()
        router_data = self.core_plugin.get_router(ctx_elevated, router_id)
        if not router_data:
            LOG.error("Couldn't read router %s data", router_id)
            return False

        # Check if the FWaaS driver supports this router
        if not self.internal_driver.should_apply_firewall_to_router(
            router_data):
            return False

        return True

    def update_router_firewall(self, context, nsxlib, router_id,
                               router_interfaces, nsx_router_id, section_id):
        """Rewrite all the FWaaS v1 rules in the router edge firewall

        This method should be called on FWaaS updates, and on router
        interfaces changes.
        """
        fw_rules = []
        fw_id = None
        if self.should_apply_firewall_to_router(context, router_id):
            # Find the firewall attached to this router
            # (must have one since should_apply returned true)
            firewall = self.get_router_firewall(context, router_id)
            fw_id = firewall['id']

            # Add the FW rules
            fw_rules.extend(self.internal_driver.get_router_translated_rules(
                router_id, firewall))

            # Add plugin additional allow rules
            fw_rules.extend(self.core_plugin.get_extra_fw_rules(
                context, router_id))

            # Add the default drop all rule
            fw_rules.append(self.internal_driver.get_default_backend_rule(
                section_id, allow_all=False))
        else:
            # default allow all rule
            fw_rules.append(self.internal_driver.get_default_backend_rule(
                section_id, allow_all=True))

        # update the backend
        nsxlib.firewall_section.update(section_id, rules=fw_rules)

        # Also update the router tags
        self.internal_driver.update_nsx_router_tags(nsx_router_id, fw_id=fw_id)

    def delete_port(self, context, port_id):
        # nothing to do in FWaaS v1
        pass



# Copyright 2015 VMware, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib.db import model_base
import sqlalchemy as sa
from sqlalchemy import orm

from neutron.db.models import l3 as l3_db
from neutron.db import models_v2
from oslo_db.sqlalchemy import models

from vmware_nsx.common import nsxv_constants


class NsxvRouterBinding(model_base.BASEV2, model_base.HasStatusDescription,
                        models.TimestampMixin):
    """Represents the mapping between neutron router and vShield Edge."""

    __tablename__ = 'nsxv_router_bindings'

    # no ForeignKey to routers.id because for now, a router can be removed
    # from routers when delete_router is executed, but the binding is only
    # removed after the Edge is deleted
    router_id = sa.Column(sa.String(36),
                          primary_key=True)
    edge_id = sa.Column(sa.String(36),
                        nullable=True)
    lswitch_id = sa.Column(sa.String(36),
                           nullable=True)
    appliance_size = sa.Column(sa.Enum(
        nsxv_constants.COMPACT,
        nsxv_constants.LARGE,
        nsxv_constants.XLARGE,
        nsxv_constants.QUADLARGE,
        name='nsxv_router_bindings_appliance_size'))
    edge_type = sa.Column(sa.Enum(nsxv_constants.SERVICE_EDGE,
                                  nsxv_constants.VDR_EDGE,
                                  name='nsxv_router_bindings_edge_type'))
    availability_zone = sa.Column(sa.String(36),
                                  nullable=True)


class NsxvEdgeVnicBinding(model_base.BASEV2, models.TimestampMixin):
    """Represents mapping between vShield Edge vnic and neutron netowrk."""

    __tablename__ = 'nsxv_edge_vnic_bindings'

    edge_id = sa.Column(sa.String(36),
                        primary_key=True)
    vnic_index = sa.Column(sa.Integer(),
                           primary_key=True)
    tunnel_index = sa.Column(sa.Integer(),
                             primary_key=True)
    network_id = sa.Column(sa.String(36), nullable=True)


class NsxvEdgeDhcpStaticBinding(model_base.BASEV2, models.TimestampMixin):
    """Represents mapping between mac addr and bindingId."""

    __tablename__ = 'nsxv_edge_dhcp_static_bindings'

    edge_id = sa.Column(sa.String(36), primary_key=True)
    mac_address = sa.Column(sa.String(32), primary_key=True)
    binding_id = sa.Column(sa.String(36), nullable=False)


class NsxvInternalNetworks(model_base.BASEV2, models.TimestampMixin):
    """Represents internal networks between NSXV plugin elements."""
    __tablename__ = 'nsxv_internal_networks'

    network_purpose = sa.Column(
        sa.Enum(nsxv_constants.INTER_EDGE_PURPOSE,
                name='nsxv_internal_networks_purpose'),
        primary_key=True)
    network_id = sa.Column(sa.String(36),
                           sa.ForeignKey("networks.id", ondelete="CASCADE"),
                           nullable=True)
    availability_zone = sa.Column(sa.String(36),
                                  primary_key=True)


class NsxvInternalEdges(model_base.BASEV2, models.TimestampMixin):
    """Represents internal Edge appliances for NSXV plugin operations."""
    __tablename__ = 'nsxv_internal_edges'

    ext_ip_address = sa.Column(sa.String(64), primary_key=True)
    router_id = sa.Column(sa.String(36), nullable=True)
    purpose = sa.Column(
        sa.Enum(nsxv_constants.INTER_EDGE_PURPOSE,
                name='nsxv_internal_edges_purpose'))


class NsxvSecurityGroupSectionMapping(model_base.BASEV2,
                                      models.TimestampMixin):
    """Backend mappings for Neutron Rule Sections.

    This class maps a neutron security group identifier to the corresponding
    NSX layer 3 section.
    """

    __tablename__ = 'nsxv_security_group_section_mappings'
    neutron_id = sa.Column(sa.String(36),
                           sa.ForeignKey('securitygroups.id',
                                         ondelete="CASCADE"),
                           primary_key=True)
    ip_section_id = sa.Column(sa.String(100))


class NsxvRuleMapping(model_base.BASEV2, models.TimestampMixin):
    """Backend mappings for Neutron Rule Sections.

    This class maps a neutron security group identifier to the corresponding
    NSX layer 3 and layer 2 sections.
    """

    __tablename__ = 'nsxv_rule_mappings'

    neutron_id = sa.Column(sa.String(36),
                           sa.ForeignKey('securitygrouprules.id',
                                         ondelete="CASCADE"),
                           primary_key=True)
    nsx_rule_id = sa.Column(sa.String(36), primary_key=True)


class NsxvPortVnicMapping(model_base.BASEV2, models.TimestampMixin):
    """Maps neutron port to NSXv VM Vnic Id."""

    __tablename__ = 'nsxv_port_vnic_mappings'

    neutron_id = sa.Column(sa.String(36),
                           sa.ForeignKey('ports.id', ondelete="CASCADE"),
                           primary_key=True)
    nsx_id = sa.Column(sa.String(42), primary_key=True)


class NsxvRouterExtAttributes(model_base.BASEV2, models.TimestampMixin):
    """Router attributes managed by NSX plugin extensions."""

    __tablename__ = 'nsxv_router_ext_attributes'

    router_id = sa.Column(sa.String(36),
                          sa.ForeignKey('routers.id', ondelete="CASCADE"),
                          primary_key=True)
    distributed = sa.Column(sa.Boolean, default=False, nullable=False)
    router_type = sa.Column(
        sa.Enum('shared', 'exclusive',
                name='nsxv_router_type'),
        default='exclusive', nullable=False)
    service_router = sa.Column(sa.Boolean, default=False, nullable=False)
    # Add a relationship to the Router model in order to instruct
    # SQLAlchemy to eagerly load this association
    router = orm.relationship(
        l3_db.Router,
        backref=orm.backref("nsx_attributes", lazy='joined',
                            uselist=False, cascade='delete'))


class NsxvTzNetworkBinding(model_base.BASEV2, models.TimestampMixin):
    """Represents a binding of a virtual network with a transport zone.

    This model class associates a Neutron network with a transport zone;
    optionally a vlan ID might be used if the binding type is 'bridge'
    """

    __tablename__ = 'nsxv_tz_network_bindings'

    network_id = sa.Column(sa.String(36),
                           sa.ForeignKey('networks.id', ondelete="CASCADE"),
                           primary_key=True)
    binding_type = sa.Column(
        sa.Enum('flat', 'vlan', 'portgroup', 'vxlan',
                name='nsxv_tz_network_bindings_binding_type'),
        nullable=False, primary_key=True)
    phy_uuid = sa.Column(sa.String(36), primary_key=True, nullable=True)
    vlan_id = sa.Column(sa.Integer, primary_key=True, nullable=True,
                        autoincrement=False)

    def __init__(self, network_id, binding_type, phy_uuid, vlan_id):
        self.network_id = network_id
        self.binding_type = binding_type
        self.phy_uuid = phy_uuid
        self.vlan_id = vlan_id

    def __repr__(self):
        return "<NsxvTzNetworkBinding(%s,%s,%s,%s)>" % (self.network_id,
                                                        self.binding_type,
                                                        self.phy_uuid,
                                                        self.vlan_id)


class NsxvPortIndexMapping(model_base.BASEV2, models.TimestampMixin):
    """Associates attached Neutron ports with the instance VNic index."""

    __tablename__ = 'nsxv_port_index_mappings'

    port_id = sa.Column(sa.String(36),
                        sa.ForeignKey('ports.id', ondelete="CASCADE"),
                        primary_key=True)
    device_id = sa.Column(sa.String(255), nullable=False)
    index = sa.Column(sa.Integer, nullable=False)
    __table_args__ = (sa.UniqueConstraint(device_id, index),
                      model_base.BASEV2.__table_args__)

    # Add a relationship to the Port model in order to instruct SQLAlchemy to
    # eagerly read port vnic-index
    port = orm.relationship(
        models_v2.Port,
        backref=orm.backref("vnic_index", lazy='joined',
                            uselist=False, cascade='delete'))


class NsxvEdgeFirewallRuleBinding(model_base.BASEV2, models.TimestampMixin):
    """Mapping between firewall rule and edge firewall rule_id."""

    __tablename__ = 'nsxv_firewall_rule_bindings'

    rule_id = sa.Column(sa.String(36),
                        primary_key=True)
    edge_id = sa.Column(sa.String(36), primary_key=True)
    rule_vse_id = sa.Column(sa.String(36))


class NsxvSpoofGuardPolicyNetworkMapping(model_base.BASEV2,
                                         models.TimestampMixin):
    """Mapping between SpoofGuard and neutron networks"""

    __tablename__ = 'nsxv_spoofguard_policy_network_mappings'

    network_id = sa.Column(sa.String(36),
                           sa.ForeignKey('networks.id', ondelete='CASCADE'),
                           primary_key=True,
                           nullable=False)
    policy_id = sa.Column(sa.String(36), nullable=False)


class NsxvLbaasLoadbalancerBinding(model_base.BASEV2, models.TimestampMixin):
    """Mapping between Edge LB and LBaaSv2"""

    __tablename__ = 'nsxv_lbaas_loadbalancer_bindings'

    loadbalancer_id = sa.Column(sa.String(36),
                                sa.ForeignKey('lbaas_loadbalancers.id',
                                              name='fk_lbaas_loadbalancers_id',
                                              ondelete="CASCADE"),
                                primary_key=True)
    edge_id = sa.Column(sa.String(36), nullable=False)
    edge_fw_rule_id = sa.Column(sa.String(36), nullable=False)
    vip_address = sa.Column(sa.String(36), nullable=False)


class NsxvLbaasListenerBinding(model_base.BASEV2, models.TimestampMixin):
    """Mapping between Edge VSE and LBaaSv2"""

    __tablename__ = 'nsxv_lbaas_listener_bindings'

    loadbalancer_id = sa.Column(sa.String(36), primary_key=True)
    listener_id = sa.Column(sa.String(36),
                            sa.ForeignKey('lbaas_listeners.id',
                                          name='fk_lbaas_listeners_id',
                                          ondelete="CASCADE"),
                            primary_key=True)
    app_profile_id = sa.Column(sa.String(36), nullable=False)
    vse_id = sa.Column(sa.String(36), nullable=False)


class NsxvLbaasPoolBinding(model_base.BASEV2, models.TimestampMixin):
    """Mapping between Edge Pool and LBaaSv2"""

    __tablename__ = 'nsxv_lbaas_pool_bindings'

    loadbalancer_id = sa.Column(sa.String(36), primary_key=True)
    pool_id = sa.Column(sa.String(36),
                        sa.ForeignKey('lbaas_pools.id',
                                      name='fk_lbaas_pools_id',
                                      ondelete="CASCADE"),
                        primary_key=True)
    edge_pool_id = sa.Column(sa.String(36), nullable=False)


class NsxvLbaasMonitorBinding(model_base.BASEV2, models.TimestampMixin):
    """Mapping between Edge Monitor and LBaaSv2"""

    __tablename__ = 'nsxv_lbaas_monitor_bindings'

    loadbalancer_id = sa.Column(sa.String(36), primary_key=True)
    pool_id = sa.Column(sa.String(36), primary_key=True)
    hm_id = sa.Column(sa.String(36),
                      sa.ForeignKey('lbaas_healthmonitors.id',
                                    name='fk_lbaas_healthmonitors_id',
                                    ondelete="CASCADE"),
                      primary_key=True)
    edge_id = sa.Column(sa.String(36), primary_key=True)
    edge_mon_id = sa.Column(sa.String(36), nullable=False)


class NsxvLbaasCertificateBinding(model_base.BASEV2, models.TimestampMixin):
    """Mapping between Edge certificate and LBaaSv2 object"""

    __tablename__ = 'nsxv_lbaas_certificate_bindings'

    cert_id = sa.Column(sa.String(128), primary_key=True)
    edge_id = sa.Column(sa.String(36), primary_key=True)
    edge_cert_id = sa.Column(sa.String(36), nullable=False)


class NsxvLbaasL7PolicyBinding(model_base.BASEV2, models.TimestampMixin):
    """Mapping between NSX Edge and LBaaSv2 L7 policy """

    __tablename__ = 'nsxv_lbaas_l7policy_bindings'

    policy_id = sa.Column(sa.String(36),
                          sa.ForeignKey('lbaas_l7policies.id',
                                        name='fk_lbaas_l7policies_id',
                                        ondelete="CASCADE"),
                          primary_key=True)
    edge_id = sa.Column(sa.String(36), nullable=False)
    edge_app_rule_id = sa.Column(sa.String(36), nullable=False)


class NsxvSubnetExtAttributes(model_base.BASEV2, models.TimestampMixin):
    """Subnet attributes managed by NSX plugin extensions."""

    __tablename__ = 'nsxv_subnet_ext_attributes'

    subnet_id = sa.Column(sa.String(36),
                          sa.ForeignKey('subnets.id', ondelete="CASCADE"),
                          primary_key=True)
    dns_search_domain = sa.Column(sa.String(255), nullable=True)
    dhcp_mtu = sa.Column(sa.Integer, nullable=True)
    # Add a relationship to the Subnet model in order to instruct
    # SQLAlchemy to eagerly load this association
    subnet = orm.relationship(
        models_v2.Subnet,
        backref=orm.backref("nsxv_subnet_attributes", lazy='joined',
                            uselist=False, cascade='delete'))


class NsxvPortExtAttributes(model_base.BASEV2, models.TimestampMixin):
    """Port attributes managed by NSX plugin extensions."""

    __tablename__ = 'nsxv_port_ext_attributes'

    port_id = sa.Column(sa.String(36),
                        sa.ForeignKey('ports.id', ondelete="CASCADE"),
                        primary_key=True)
    vnic_type = sa.Column(sa.String(64), nullable=False,
                          default=portbindings.VNIC_NORMAL,
                          server_default=portbindings.VNIC_NORMAL)
    # Add a relationship to the port model in order to instruct
    # SQLAlchemy to eagerly load this association
    port = orm.relationship(
        models_v2.Port,
        backref=orm.backref("nsx_port_attributes", lazy='joined',
                            uselist=False, cascade='delete'))


class NsxvBgpSpeakerBinding(model_base.BASEV2, models.TimestampMixin):
    # Maps bgp_speaker_id to NSXv edge id
    __tablename__ = 'nsxv_bgp_speaker_bindings'

    edge_id = sa.Column(sa.String(36), primary_key=True)
    bgp_speaker_id = sa.Column(sa.String(36),
                               sa.ForeignKey('bgp_speakers.id',
                                             ondelete='CASCADE'),
                               nullable=False)
    # A given BGP speaker sets the value of its BGP Identifier to an IP address
    # that is assigned to that BGP speaker.
    bgp_identifier = sa.Column(sa.String(64), nullable=False)


class NsxvBgpPeerEdgeBinding(model_base.BASEV2, models.TimestampMixin):
    # Maps between bgp-peer and edges service gateway.
    __tablename__ = 'nsxv_bgp_peer_edge_bindings'
    peer_id = sa.Column(sa.String(36),
                        sa.ForeignKey('bgp_peers.id',
                                      ondelete='CASCADE'),
                        primary_key=True,
                        nullable=False)
    edge_id = sa.Column(sa.String(36), nullable=False)



# Copyright (c) 2014 VMware.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from neutron_lib import context
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.tests import base
import neutron.tests.unit.db.test_db_base_plugin_v2 as test_plugin
from neutron_lib.api.definitions import portbindings
from neutron_lib import exceptions as exp
from neutron_lib.plugins import directory
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.db import db as nsx_db
from vmware_nsx.dvs import dvs
from vmware_nsx.dvs import dvs_utils

PLUGIN_NAME = 'vmware_nsx.plugin.NsxDvsPlugin'


class fake_session(object):
    def __init__(self, *ret):
        self._vim = mock.Mock()

    def invoke_api(self, *args, **kwargs):
        pass

    def wait_for_task(self, task):
        pass

    def vim(self):
        return self._vim


class DvsTestCase(base.BaseTestCase):

    @mock.patch.object(dvs_utils, 'dvs_create_session',
                       return_value=fake_session())
    @mock.patch.object(dvs.SingleDvsManager, '_get_dvs_moref_by_name',
                       return_value=mock.MagicMock())
    def setUp(self, mock_moref, mock_session):
        super(DvsTestCase, self).setUp()
        cfg.CONF.set_override('dvs_name', 'fake_dvs', group='dvs')
        self._dvs = dvs.SingleDvsManager()
        self.assertEqual(mock_moref.return_value, self._dvs._dvs_moref)
        mock_moref.assert_called_once_with(mock_session.return_value,
                                           'fake_dvs')

    @mock.patch.object(dvs_utils, 'dvs_create_session',
                       return_value=fake_session())
    def test_dvs_not_found(self, mock_session):
        self.assertRaises(nsx_exc.DvsNotFound,
                          dvs.SingleDvsManager)

    @mock.patch.object(dvs.DvsManager, '_get_port_group_spec',
                       return_value='fake-spec')
    def test_add_port_group(self, fake_get_spec):
        self._dvs.add_port_group('fake-uuid', vlan_tag=7)
        fake_get_spec.assert_called_once_with('fake-uuid', 7, trunk_mode=False)

    @mock.patch.object(dvs.DvsManager, '_get_port_group_spec',
                       return_value='fake-spec')
    def test_add_port_group_with_exception(self, fake_get_spec):
        with (
            mock.patch.object(self._dvs._dvs._session, 'wait_for_task',
                              side_effect=exp.NeutronException())
        ):
            self.assertRaises(exp.NeutronException,
                              self._dvs.add_port_group,
                              'fake-uuid', 7,
                              trunk_mode=False)
            fake_get_spec.assert_called_once_with('fake-uuid', 7,
                                                  trunk_mode=False)

    @mock.patch.object(dvs.DvsManager, '_net_id_to_moref',
                       return_value='fake-moref')
    def test_delete_port_group(self, fake_get_moref):
        self._dvs.delete_port_group('fake-uuid')
        fake_get_moref.assert_called_once_with(mock.ANY, 'fake-uuid')

    @mock.patch.object(dvs.DvsManager, '_net_id_to_moref',
                       return_value='fake-moref')
    def test_delete_port_group_with_exception(self, fake_get_moref):
        with (
            mock.patch.object(self._dvs._dvs._session, 'wait_for_task',
                              side_effect=exp.NeutronException())
        ):
            self.assertRaises(exp.NeutronException,
                              self._dvs.delete_port_group,
                              'fake-uuid')
            fake_get_moref.assert_called_once_with(mock.ANY, 'fake-uuid')

    @mock.patch.object(dvs.DvsManager, '_update_vxlan_port_groups_config')
    @mock.patch.object(dvs.DvsManager, '_get_port_group_spec',
                       return_value='fake-spec')
    @mock.patch.object(dvs.DvsManager, '_net_id_to_moref',
                       return_value='fake-moref')
    def test_update_vxlan_net_group_conf(self, fake_get_moref,
                                         fake_get_spec, fake_update_vxlan):
        net_id = 'vxlan-uuid'
        vlan = 7
        self._dvs.add_port_group(net_id, vlan)
        self._dvs.net_id_to_moref(net_id)
        fake_get_moref.assert_called_once_with(mock.ANY, net_id)
        fake_get_spec.assert_called_once_with(net_id, vlan, trunk_mode=False)

    @mock.patch.object(dvs.DvsManager, '_update_net_port_groups_config')
    @mock.patch.object(dvs.DvsManager, '_get_port_group_spec',
                       return_value='fake-spec')
    @mock.patch.object(dvs.DvsManager, '_net_id_to_moref',
                       return_value='dvportgroup-fake-moref')
    def test_update_flat_net_conf(self, fake_get_moref,
                                  fake_get_spec, fake_update_net):
        net_id = 'flat-uuid'
        vlan = 7
        self._dvs.add_port_group(net_id, vlan)
        self._dvs.net_id_to_moref(net_id)
        fake_get_moref.assert_called_once_with(mock.ANY, net_id)
        fake_get_spec.assert_called_once_with(net_id, vlan, trunk_mode=False)


class NeutronSimpleDvsTestCase(test_plugin.NeutronDbPluginV2TestCase):

    @mock.patch.object(dvs_utils, 'dvs_create_session',
                       return_value=fake_session())
    @mock.patch.object(dvs.SingleDvsManager, '_get_dvs_moref_by_name',
                       return_value=mock.MagicMock())
    def setUp(self, mock_moref, mock_session,
              plugin=PLUGIN_NAME,
              ext_mgr=None,
              service_plugins=None):
        # Ensure that DVS is enabled
        cfg.CONF.set_override('host_ip', 'fake_ip', group='dvs')
        cfg.CONF.set_override('host_username', 'fake_user', group='dvs')
        cfg.CONF.set_override('host_password', 'fake_password', group='dvs')
        cfg.CONF.set_override('dvs_name', 'fake_dvs', group='dvs')
        super(NeutronSimpleDvsTestCase, self).setUp(plugin=plugin)
        self._plugin = directory.get_plugin()


class NeutronSimpleDvsTest(NeutronSimpleDvsTestCase):

    def _create_and_delete_dvs_network(self, network_type='flat', vlan_tag=0,
                                       trunk_mode=False):
        params = {'provider:network_type': network_type,
                  'provider:physical_network': 'fake-moid',
                  'name': 'fake-name'}
        if network_type == 'vlan':
            params['provider:segmentation_id'] = vlan_tag
        if trunk_mode:
            params['vlan_transparent'] = True
        params['arg_list'] = tuple(params.keys())
        with mock.patch.object(self._plugin._dvs,
                               'add_port_group') as mock_add,\
                mock.patch.object(self._plugin._dvs,
                    'delete_port_group') as mock_delete,\
                mock.patch.object(dvs.DvsManager,
                    '_get_trunk_vlan_spec') as mock_trunk_vlan:
            with self.network(**params) as network:
                ctx = context.get_admin_context()
                id = network['network']['id']
                dvs_id = '%s-%s' % (network['network']['name'], id)
                binding = nsx_db.get_network_bindings(ctx.session, id)
                self.assertIsNotNone(binding)
                if network_type == 'flat':
                    self.assertEqual('flat', binding[0].binding_type)
                    self.assertEqual(0, binding[0].vlan_id)
                    self.assertEqual('dvs', binding[0].phy_uuid)
                elif network_type == 'vlan':
                    self.assertEqual('vlan', binding[0].binding_type)
                    self.assertEqual(vlan_tag, binding[0].vlan_id)
                    self.assertEqual('dvs', binding[0].phy_uuid)
                elif network_type == 'portgroup':
                    self.assertEqual('portgroup', binding[0].binding_type)
                    self.assertEqual(0, binding[0].vlan_id)
                    self.assertEqual('fake-moid', binding[0].phy_uuid)
                else:
                    self.fail()
            if network_type != 'portgroup':
                mock_add.assert_called_once_with(dvs_id, vlan_tag,
                                                 trunk_mode=trunk_mode)
            else:
                mock_add.call_count = 0
                mock_delete.call_count = 0
            if trunk_mode:
                mock_trunk_vlan.called_once_with(start=0, end=4094)
            else:
                mock_trunk_vlan.call_count = 0

    def test_create_and_delete_dvs_network_tag(self):
        self._create_and_delete_dvs_network(network_type='vlan', vlan_tag=7)

    def test_create_and_delete_dvs_network_flat(self):
        self._create_and_delete_dvs_network()

    def test_create_and_delete_dvs_network_flat_vlan_transparent(self):
        self._create_and_delete_dvs_network(trunk_mode=True)

    @mock.patch.object(dvs.DvsManager, 'get_port_group_info')
    @mock.patch.object(dvs.DvsManager, '_net_id_to_moref')
    def test_create_and_delete_dvs_network_portgroup(self, fake_get_moref,
                                                     fake_pg_info):
        fake_pg_info.return_value = {'name': 'fake-name'}
        self._create_and_delete_dvs_network(network_type='portgroup')
        self.assertTrue(fake_get_moref.call_count)
        self.assertTrue(fake_pg_info.call_count)

    @mock.patch.object(dvs.DvsManager, 'get_port_group_info')
    @mock.patch.object(dvs.DvsManager, '_net_id_to_moref')
    def test_create_and_delete_dvs_network_portgroup_vlan(self,
                                                          fake_get_moref,
                                                          fake_pg_info):
        fake_pg_info.return_value = {'name': 'fake-name'}
        self._create_and_delete_dvs_network(network_type='portgroup',
                                            vlan_tag=7)
        self.assertTrue(fake_get_moref.call_count)
        self.assertTrue(fake_pg_info.call_count)

    def test_create_and_delete_dvs_port(self):
        params = {'provider:network_type': 'vlan',
                  'provider:physical_network': 'dvs',
                  'provider:segmentation_id': 7}
        params['arg_list'] = tuple(params.keys())
        with mock.patch.object(self._plugin._dvs, 'add_port_group'),\
                mock.patch.object(self._plugin._dvs, 'delete_port_group'):
            with self.network(**params) as network,\
                    self.subnet(network) as subnet,\
                    self.port(subnet) as port:
                self.assertEqual('dvs',
                                 port['port'][portbindings.VIF_TYPE])
                port_status = port['port']['status']
                self.assertEqual(port_status, 'ACTIVE')

    def test_create_router_only_dvs_backend(self):
        data = {'router': {'tenant_id': 'whatever'}}
        data['router']['name'] = 'router1'
        data['router']['external_gateway_info'] = {'network_id': 'whatever'}
        self.assertRaises(exp.BadRequest,
                          self._plugin.create_router,
                          context.get_admin_context(),
                          data)

    def test_dvs_get_id(self):
        id = uuidutils.generate_uuid()
        net = {'name': '',
               'id': id}
        expected = id
        self.assertEqual(expected, self._plugin._dvs_get_id(net))
        net = {'name': 'pele',
               'id': id}
        expected = '%s-%s' % ('pele', id)
        self.assertEqual(expected, self._plugin._dvs_get_id(net))
        name = 'X' * 500
        net = {'name': name,
               'id': id}
        expected = '%s-%s' % (name[:43], id)
        self.assertEqual(expected, self._plugin._dvs_get_id(net))

    def test_update_dvs_network(self):
        """Test update of a DVS network
        """
        params = {'provider:network_type': 'flat',
                  'admin_state_up': True,
                  'name': 'test_net',
                  'tenant_id': 'fake_tenant',
                  'shared': False,
                  'port_security_enabled': False}

        with mock.patch.object(self._plugin._dvs, 'add_port_group'):
            ctx = context.get_admin_context()
            # create the initial network
            network = self._plugin.create_network(ctx, {'network': params})
            id = network['id']

            # update the different attributes of the DVS network

            # cannot update the provider type
            self.assertRaises(
                exp.InvalidInput,
                self._plugin.update_network,
                ctx, id,
                {'network': {'provider:network_type': 'vlan'}})

            # update the Shared attribute
            self.assertEqual(False, network['shared'])
            updated_net = self._plugin.update_network(
                ctx, id,
                {'network': {'shared': True}})
            self.assertEqual(True, updated_net['shared'])

            # Update the description attribute
            self.assertIsNone(network['description'])
            updated_net = self._plugin.update_network(
                ctx, id,
                {'network': {'description': 'test'}})
            self.assertEqual('test', updated_net['description'])

            # update the port security attribute
            self.assertEqual(False, network['port_security_enabled'])
            updated_net = self._plugin.update_network(
                ctx, id,
                {'network': {'port_security_enabled': True}})
            self.assertEqual(True, updated_net['port_security_enabled'])

    @mock.patch.object(dvs.DvsManager, 'get_port_group_info')
    @mock.patch.object(dvs.DvsManager, '_net_id_to_moref')
    def test_create_and_delete_portgroup_network_invalid_name(self,
                                                          fake_get_moref,
                                                          fake_pg_info):
        fake_pg_info.return_value = {'name': 'fake-different-name'}
        data = {'network': {'provider:network_type': 'portgroup',
                            'name': 'fake-name',
                            'admin_state_up': True}}
        self.assertRaises(exp.BadRequest, self._plugin.create_network,
                          context.get_admin_context(), data)



# Copyright 2016 VMware, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from osc_lib import utils as osc_utils


cached_extensions = None


def get_extensions(client_manager):
    """Return a list of all current extensions aliases"""
    # Return previously calculated results
    global cached_extensions
    if cached_extensions is not None:
        return cached_extensions
    # Get supported extensions from the manager
    data = client_manager.network.extensions()
    extensions = []
    for s in data:
        prop = osc_utils.get_item_properties(
            s, ('Alias',), formatters={})
        extensions.append(prop[0])
    # Save the results in the global cache
    cached_extensions = extensions
    return extensions



# Copyright 2013 VMware, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
import sqlalchemy as sa

from oslo_db.sqlalchemy import models


class VcnsRouterBinding(model_base.BASEV2, model_base.HasStatusDescription,
                        models.TimestampMixin):
    """Represents the mapping between neutron router and vShield Edge."""

    __tablename__ = 'vcns_router_bindings'

    # no ForeignKey to routers.id because for now, a router can be removed
    # from routers when delete_router is executed, but the binding is only
    # removed after the Edge is deleted
    router_id = sa.Column(sa.String(36),
                          primary_key=True)
    edge_id = sa.Column(sa.String(16),
                        nullable=True)
    lswitch_id = sa.Column(sa.String(36),
                           nullable=False)



# Copyright 2013 OpenStack Foundation.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import time

import eventlet
import mock

from vmware_nsx.api_client import client as nsx_client
from vmware_nsx.api_client import eventlet_client
from vmware_nsx import extensions
import vmware_nsx.plugin as neutron_plugin
from vmware_nsx.plugins.nsx_v.vshield.common import (
    VcnsApiClient as vcnsapi)
from vmware_nsx.plugins.nsx_v.vshield import edge_utils
from vmware_nsx.plugins.nsx_v.vshield import vcns
import vmware_nsx.plugins.nsx_v.vshield.vcns_driver as vcnsdriver


plugin = neutron_plugin.NsxPlugin
api_client = nsx_client.NsxApiClient
evt_client = eventlet_client.EventletApiClient
vcns_class = vcns.Vcns
vcns_driver = vcnsdriver.VcnsDriver
vcns_api_helper = vcnsapi.VcnsApiHelper
edge_manage_class = edge_utils.EdgeManager

STUBS_PATH = os.path.join(os.path.dirname(__file__), 'etc')
NSXEXT_PATH = os.path.dirname(extensions.__file__)
NSXAPI_NAME = '%s.%s' % (api_client.__module__, api_client.__name__)
PLUGIN_NAME = '%s.%s' % (plugin.__module__, plugin.__name__)
CLIENT_NAME = '%s.%s' % (evt_client.__module__, evt_client.__name__)
VCNS_NAME = '%s.%s' % (vcns_class.__module__, vcns_class.__name__)
VCNS_DRIVER_NAME = '%s.%s' % (vcns_driver.__module__, vcns_driver.__name__)
VCNSAPI_NAME = '%s.%s' % (vcns_api_helper.__module__, vcns_api_helper.__name__)
EDGE_MANAGE_NAME = '%s.%s' % (edge_manage_class.__module__,
                              edge_manage_class.__name__)

# Mock for the tenacity retrying sleeping method
eventlet.monkey_patch()
mocked_retry_sleep = mock.patch.object(time, 'sleep')
mocked_retry_sleep.start()


def get_fake_conf(filename):
    return os.path.join(STUBS_PATH, filename)


def nsx_method(method_name, module_name='nsxlib'):
    return '%s.%s.%s' % ('vmware_nsx', module_name,
                         method_name)



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from vmware_nsx.extensions import projectpluginmap
from vmware_nsx.services.fwaas.common import fwaas_callbacks_v1 as com_clbcks
from vmware_nsx.services.fwaas.nsx_tv import edge_fwaas_driver_v1 as tv_driver

LOG = logging.getLogger(__name__)


class NsxvFwaasCallbacks(com_clbcks.NsxFwaasCallbacks):
    """NSX-V RPC callbacks for Firewall As A Service - V1."""

    def __init__(self):
        super(NsxvFwaasCallbacks, self).__init__()
        # update the fwaas driver in case of TV plugin
        if self.fwaas_enabled:
            if self.fwaas_driver.driver_name == tv_driver.FWAAS_DRIVER_NAME:
                self.internal_driver = self.fwaas_driver.get_V_driver()
            else:
                self.internal_driver = self.fwaas_driver

    @property
    def plugin_type(self):
        return projectpluginmap.NsxPlugins.NSX_V

    def should_apply_firewall_to_router(self, context, router, router_id):
        """Return True if the FWaaS rules should be added to this router."""
        # in case of a distributed-router:
        # router['id'] is the id of the neutron router (=tlr)
        # and router_id is the plr/tlr (the one that is being updated)
        if not super(NsxvFwaasCallbacks, self).should_apply_firewall_to_router(
            context, router['id']):
            return False

        # get all the relevant router info
        # ("router" does not have all the fields)
        ctx_elevated = context.elevated()
        router_data = self.core_plugin.get_router(ctx_elevated, router['id'])
        if not router_data:
            LOG.error("Couldn't read router %s data", router['id'])
            return False

        if router_data.get('distributed'):
            if router_id == router['id']:
                # Do not add firewall rules on the tlr router.
                return False

        # Check if the FWaaS driver supports this router
        if not self.internal_driver.should_apply_firewall_to_router(
            router_data, raise_exception=False):
            return False

        return True

    def get_fwaas_rules_for_router(self, context, router_id):
        """Return the list of (translated) FWaaS rules for this router."""
        ctx_elevated = context.elevated()
        fw_id = self._get_router_firewall_id(ctx_elevated, router_id)
        if fw_id:
            return self._get_fw_applicable_rules(ctx_elevated, fw_id)
        return []

    def _get_fw_applicable_rules(self, context, fw_id):
        fw = self._get_fw_from_plugin(context, fw_id)
        if fw is not None and fw['id'] == fw_id:
            return self.internal_driver.get_firewall_translated_rules(fw)
        return []



# Copyright 2016 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""NSXv3 add nsx_service_bindings and nsx_dhcp_bindings tables

Revision ID: c644ec62c585
Revises: c288bb6a7252
Create Date: 2016-04-29 23:19:39.523196

"""

# revision identifiers, used by Alembic.
revision = 'c644ec62c585'
down_revision = 'c288bb6a7252'

from alembic import op
import sqlalchemy as sa

from vmware_nsxlib.v3 import nsx_constants


nsx_service_type_enum = sa.Enum(
    nsx_constants.SERVICE_DHCP,
    name='neutron_nsx_service_bindings_service_type')


def upgrade():
    op.create_table(
        'neutron_nsx_service_bindings',
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.Column('port_id', sa.String(36), nullable=True),
        sa.Column('nsx_service_type', nsx_service_type_enum, nullable=False),
        sa.Column('nsx_service_id', sa.String(36), nullable=False),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_id', 'nsx_service_type'))

    op.create_table(
        'neutron_nsx_dhcp_bindings',
        sa.Column('port_id', sa.String(36), nullable=False),
        sa.Column('subnet_id', sa.String(36), nullable=False),
        sa.Column('ip_address', sa.String(64), nullable=False),
        sa.Column('nsx_service_id', sa.String(36), nullable=False),
        sa.Column('nsx_binding_id', sa.String(36), nullable=False),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('port_id', 'nsx_binding_id'))



# Copyright 2014 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_serialization import jsonutils
import six

from vmware_nsx.plugins.nsx_v.vshield import vcns


@six.add_metaclass(abc.ABCMeta)
class NsxvEdgeCfgObj(object):

    def __init__(self):
        return

    @abc.abstractmethod
    def get_service_name(self):
        return

    @abc.abstractmethod
    def serializable_payload(self):
        return

    @staticmethod
    def get_object(vcns_obj, edge_id, service_name):
        uri = "%s/%s/%s" % (vcns.URI_PREFIX,
                            edge_id,
                            service_name)

        h, v = vcns_obj.do_request(
            vcns.HTTP_GET,
            uri,
            decode=True)

        return v

    def submit_to_backend(self, vcns_obj, edge_id):
        uri = "%s/%s/%s/config" % (vcns.URI_PREFIX,
                                   edge_id,
                                   self.get_service_name())

        payload = jsonutils.dumps(self.serializable_payload(), sort_keys=True)

        if payload:
            return vcns_obj.do_request(
                vcns.HTTP_PUT,
                uri,
                payload,
                format='json',
                encode=False)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import itertools

import vmware_nsx.common.config
import vmware_nsx.dhcp_meta.lsnmanager
import vmware_nsx.dhcp_meta.nsx
import vmware_nsx.dvs.dvs_utils
import vmware_nsx.extensions.networkgw


def list_opts():
    return [('DEFAULT',
             itertools.chain(
                 vmware_nsx.common.config.cluster_opts,
                 vmware_nsx.common.config.connection_opts,
                 vmware_nsx.common.config.nsx_common_opts)),
            ('NSX', vmware_nsx.common.config.base_opts),
            ('NSX_SYNC', vmware_nsx.common.config.sync_opts),
            ('nsxv', vmware_nsx.common.config.nsxv_opts),
            ('nsx_v3', vmware_nsx.common.config.nsx_v3_opts),
            ('QUOTAS', vmware_nsx.extensions.networkgw.nw_gw_quota_opts),
            ('dvs', vmware_nsx.dvs.dvs_utils.dvs_opts),
            ('NSX_DHCP', vmware_nsx.dhcp_meta.nsx.dhcp_opts),
            ('NSX_METADATA', vmware_nsx.dhcp_meta.nsx.metadata_opts),
            ('NSX_LSN', vmware_nsx.dhcp_meta.lsnmanager.lsn_opts)]



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron.api.v2 import resource_helper
from neutron_lib.api import extensions
from neutron_lib import exceptions as nexception

from vmware_nsx._i18n import _


HOUSEKEEPER_RESOURCE_NAME = "housekeeper"
HOUSEKEEPERS = "housekeepers"

# The housekeeper tasks table is read only
RESOURCE_ATTRIBUTE_MAP = {
    HOUSEKEEPERS: {
        'name': {
            'allow_post': False, 'allow_put': False, 'is_visible': True},
        'description': {
            'allow_post': False, 'allow_put': False, 'is_visible': True},
        'enabled': {
            'allow_post': False, 'allow_put': False, 'is_visible': True},
    }
}


class Housekeeper(extensions.ExtensionDescriptor):
    """API extension for NSX housekeeper jobs."""

    @classmethod
    def get_name(cls):
        return "Housekeeper"

    @classmethod
    def get_alias(cls):
        return HOUSEKEEPER_RESOURCE_NAME

    @classmethod
    def get_description(cls):
        return "NSX plugin housekeeping services."

    @classmethod
    def get_updated(cls):
        return "2016-11-20T00:00:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        plural_mappings = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        member_actions = {}
        return resource_helper.build_resource_info(plural_mappings,
                                                   RESOURCE_ATTRIBUTE_MAP,
                                                   None,
                                                   action_map=member_actions,
                                                   register_quota=True,
                                                   translate_name=True)

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}


class HousekeeperReadOnly(nexception.NotAuthorized):
    message = _("NSX housekeeper tasks are read-only.")


class HousekeeperPluginBase(object):

    @abc.abstractmethod
    def create_housekeeper(self, context, housekeeper):
        raise HousekeeperReadOnly()

    @abc.abstractmethod
    def update_housekeeper(self, context, name, housekeeper):
        pass

    @abc.abstractmethod
    def get_housekeeper(self, context, name, fields=None):
        pass

    @abc.abstractmethod
    def delete_housekeeper(self, context, name):
        raise HousekeeperReadOnly()

    @abc.abstractmethod
    def get_housekeepers(self, context, filters=None, fields=None,
                         sorts=None, limit=None, marker=None,
                         page_reverse=False):
        pass



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import constants as plugin_const
from neutron_lib.plugins import directory
from oslo_log import log as logging

from vmware_nsx.extensions import projectpluginmap

LOG = logging.getLogger(__name__)


class LoadbalancerBaseManager(object):
    _lbv2_driver = None
    _core_plugin = None
    _flavor_plugin = None

    def __init__(self):
        super(LoadbalancerBaseManager, self).__init__()

    def _get_plugin(self, plugin_type):
        return directory.get_plugin(plugin_type)

    @property
    def lbv2_driver(self):
        if not self._lbv2_driver:
            plugin = self._get_plugin(
                plugin_const.LOADBALANCERV2)
            self._lbv2_driver = (
                plugin.drivers['vmwareedge'])

        return self._lbv2_driver

    @property
    def core_plugin(self):
        if not self._core_plugin:
            self._core_plugin = (
                self._get_plugin(plugin_const.CORE))

        return self._core_plugin

    @property
    def flavor_plugin(self):
        if not self._flavor_plugin:
            self._flavor_plugin = (
                self._get_plugin(plugin_const.FLAVORS))

        return self._flavor_plugin


class EdgeLoadbalancerBaseManager(LoadbalancerBaseManager):

    def __init__(self, vcns_driver):
        super(EdgeLoadbalancerBaseManager, self).__init__()
        self.vcns_driver = vcns_driver

    @property
    def vcns(self):
        return self.vcns_driver.vcns

    @property
    def core_plugin(self):
        if not self._core_plugin:
            self._core_plugin = (
                self._get_plugin(plugin_const.CORE))
            if self._core_plugin.is_tvd_plugin():
                # get the plugin that match this driver
                self._core_plugin = self._core_plugin.get_plugin_by_type(
                    projectpluginmap.NsxPlugins.NSX_V)
        return self._core_plugin


class Nsxv3LoadbalancerBaseManager(LoadbalancerBaseManager):

    def __init__(self):
        super(Nsxv3LoadbalancerBaseManager, self).__init__()

    @property
    def core_plugin(self):
        if not self._core_plugin:
            self._core_plugin = (
                self._get_plugin(plugin_const.CORE))
            if self._core_plugin.is_tvd_plugin():
                # get the plugin that match this driver
                self._core_plugin = self._core_plugin.get_plugin_by_type(
                    projectpluginmap.NsxPlugins.NSX_T)
        return self._core_plugin
